import dataclasses
from collections.abc import Mapping
from typing import Literal

import matplotlib.pyplot as plt
import torch
import xarray as xr

from fme.core.dataset.data_typing import VariableMetadata
from fme.core.distributed import Distributed
from fme.core.gridded_ops import GriddedOperations
from fme.core.typing_ import TensorDict, TensorMapping
from fme.core.wandb import Image, WandB

from ..plotting import get_cmap_limits, plot_imshow


@dataclasses.dataclass
class _TargetGenPair:
    name: str
    target: torch.Tensor
    gen: torch.Tensor
    ops: GriddedOperations

    def bias(self):
        return self.gen - self.target

    def rmse(self) -> float:
        ret = float(
            self.ops.area_weighted_rmse(
                predicted=self.gen,
                truth=self.target,
                name=self.name,
            )
            .cpu()
            .numpy()
        )
        return ret

    def weighted_mean_bias(self) -> float:
        return float(
            self.ops.area_weighted_mean_bias(
                predicted=self.gen,
                truth=self.target,
                name=self.name,
            )
            .cpu()
            .numpy()
        )


def get_gen_shape(gen_data: TensorMapping):
    for name in gen_data:
        return gen_data[name].shape


class TimeMeanAggregator:
    _image_captions = {
        "bias_map": "{name} time-mean bias (generated - reference) [{units}]",
        "gen_map": "{name} time-mean generated [{units}]",
    }

    def __init__(
        self,
        gridded_operations: GriddedOperations,
        target: Literal["norm", "denorm"] = "denorm",
        variable_metadata: Mapping[str, VariableMetadata] | None = None,
        reference_means: xr.Dataset | None = None,
    ):
        """
        Args:
            gridded_operations: Computes gridded operations.
            target: Whether to compute metrics on the normalized or denormalized data,
                defaults to "denorm".
            variable_metadata: Mapping of variable names their metadata that will
                used in generating logged image captions.
            reference_means: Dataset containing reference time-mean values
                for bias computation.
        """
        self._ops = gridded_operations
        self._target = target
        if variable_metadata is None:
            self._variable_metadata: Mapping[str, VariableMetadata] = {}
        else:
            self._variable_metadata = variable_metadata
        # Dictionaries of tensors of shape [n_lat, n_lon] represnting time means
        self._data: TensorDict | None = None
        self._n_timesteps = 0
        self._n_samples: int | None = None
        self._reference_means = reference_means
        self._reference_validated = False

    @staticmethod
    def _add_or_initialize_time_mean(
        maybe_dict: TensorDict | None,
        new_data: TensorMapping,
        ignore_initial: bool = False,
    ) -> TensorDict:
        sample_dim = 0
        time_dim = 1
        if ignore_initial:
            time_slice = slice(1, None)
        else:
            time_slice = slice(0, None)
        if maybe_dict is None:
            d: TensorDict = {
                name: tensor[:, time_slice].sum(dim=time_dim).sum(dim=sample_dim)
                for name, tensor in new_data.items()
            }
        else:
            d = dict(maybe_dict)
            for name, tensor in new_data.items():
                d[name] += tensor[:, time_slice].sum(dim=time_dim).sum(dim=sample_dim)
        return d

    @torch.no_grad()
    def record_batch(
        self,
        data: TensorMapping,
        i_time_start: int = 0,
    ):
        ignore_initial = i_time_start == 0
        self._data = self._add_or_initialize_time_mean(self._data, data, ignore_initial)
        if self._n_samples is None:
            self._n_samples = data[list(data)[0]].size(0)
        if ignore_initial:
            self._n_timesteps = data[list(data)[0]].size(1) - 1
        else:
            self._n_timesteps += data[list(data)[0]].size(1)
        if not self._reference_validated:
            if self._reference_means is not None:
                self.get_logs(label="")
            self._reference_validated = True

    def get_data(self) -> TensorDict:
        if self._n_timesteps == 0 or self._data is None:
            raise ValueError("No data recorded.")

        ret = {}
        dist = Distributed.get_instance()
        names = sorted(list(self._data.keys()))  # sort for rank-consistent order
        for name in names:
            value = self._data[name]
            gen = dist.reduce_mean(value / self._n_timesteps / self._n_samples)
            ret[name] = gen
        return ret

    @torch.no_grad()
    def get_logs(self, label: str) -> dict[str, float | Image]:
        logs: dict[str, float | Image] = {}
        data = self.get_data()
        gen_map_key = "gen_map"
        wandb = WandB.get_instance()
        for name, pred in data.items():
            vmin_pred, vmax_pred = get_cmap_limits(pred.cpu().numpy())
            prediction_image = wandb.Image(
                plot_imshow(pred.cpu().numpy()),
                caption=self._get_caption(gen_map_key, name, vmin_pred, vmax_pred),
            )
            plt.close("all")
            logs.update(
                {
                    f"{gen_map_key}/{name}": prediction_image,
                }
            )
            if self._reference_means is not None and name in self._reference_means:
                pair = _TargetGenPair(
                    name=name,
                    target=torch.as_tensor(
                        self._reference_means[name].values, device=pred.device
                    ),
                    gen=pred,
                    ops=self._ops,
                )
                bias_map = pair.bias().cpu().numpy()
                vmin_bias, vmax_bias = get_cmap_limits(bias_map, diverging=True)
                bias_fig = plot_imshow(
                    bias_map, vmin=vmin_bias, vmax=vmax_bias, cmap="RdBu_r"
                )
                bias_image = wandb.Image(
                    bias_fig,
                    caption=self._get_caption("bias_map", name, vmin_bias, vmax_bias),
                )
                logs.update({f"ref_bias/{name}": pair.weighted_mean_bias()})
                logs.update({f"ref_rmse/{name}": pair.rmse()})
                logs.update({f"ref_bias_map/{name}": bias_image})

        if len(label) != 0:
            return {f"{label}/{key}": logs[key] for key in logs}
        return logs

    def _get_caption(self, key: str, name: str, vmin: float, vmax: float) -> str:
        if name in self._variable_metadata:
            caption_name = self._variable_metadata[name].long_name
            units = self._variable_metadata[name].units
        else:
            caption_name, units = name, "unknown_units"
        caption = self._image_captions[key].format(name=caption_name, units=units)
        caption += f" vmin={vmin:.4g}, vmax={vmax:.4g}."
        return caption

    def get_dataset(self) -> xr.Dataset:
        dims = ("lat", "lon")
        data = {}
        for name, pred in self.get_data().items():
            if name in self._variable_metadata:
                long_name = self._variable_metadata[name].long_name
                units = self._variable_metadata[name].units
            else:
                long_name = name
                units = "unknown_units"
            gen_metadata = VariableMetadata(long_name=long_name, units=units)._asdict()
            data.update(
                {
                    f"gen_map-{name}": xr.DataArray(
                        pred.cpu(),
                        dims=dims,
                        attrs=gen_metadata,
                    ),
                }
            )
        return xr.Dataset(data)


class TimeMeanEvaluatorAggregator:
    """Statistics and images on the time-mean state.

    This aggregator keeps track of the time-mean state, then computes
    statistics and images on that time-mean state when logs are retrieved.
    """

    _image_captions = {
        "bias_map": "{name} time-mean bias (generated - target) [{units}]",
        "gen_map": "{name} time-mean generated [{units}]",
    }

    def __init__(
        self,
        ops: GriddedOperations,
        horizontal_dims: list[str],
        target: Literal["norm", "denorm"] = "denorm",
        variable_metadata: Mapping[str, VariableMetadata] | None = None,
        reference_means: xr.Dataset | None = None,
        channel_mean_names: list[str] | None = None,
    ):
        """
        Args:
            ops: Computes gridded operations.
            horizontal_dims: Names of the horizontal dimensions.
            target: Whether to compute metrics on the normalized or denormalized data,
                defaults to "denorm".
            variable_metadata: Mapping of variable names their metadata that will
                used in generating logged image captions.
            reference_means: Dataset containing reference time-mean values
                for bias computation.
            channel_mean_names: Names of variables whose RMSE will be averaged. If
                not provided, all available variables will be used.
        """
        self._ops = ops
        self._horizontal_dims = horizontal_dims
        self._target = target
        self._dist = Distributed.get_instance()
        if variable_metadata is None:
            self._variable_metadata: Mapping[str, VariableMetadata] = {}
        else:
            self._variable_metadata = variable_metadata
        # Dictionaries of tensors of shape [n_lat, n_lon] represnting time means
        self._target_agg = TimeMeanAggregator(
            gridded_operations=ops, target=target, variable_metadata=variable_metadata
        )
        self._gen_agg = TimeMeanAggregator(
            gridded_operations=ops,
            target=target,
            variable_metadata=variable_metadata,
            reference_means=reference_means,
        )
        self._channel_mean_names = channel_mean_names

    @torch.no_grad()
    def record_batch(
        self,
        target_data: TensorMapping,
        gen_data: TensorMapping,
        target_data_norm: TensorMapping,
        gen_data_norm: TensorMapping,
        i_time_start: int = 0,
    ):
        if self._target == "norm":
            target_data = target_data_norm
            gen_data = gen_data_norm
        self._target_agg.record_batch(target_data, i_time_start)
        self._gen_agg.record_batch(gen_data, i_time_start)

    def _get_target_gen_pairs(self) -> list[_TargetGenPair]:
        target_data = self._target_agg.get_data()
        gen_data = self._gen_agg.get_data()

        ret = []
        for name in gen_data.keys():
            ret.append(
                _TargetGenPair(
                    gen=gen_data[name],
                    target=target_data[name],
                    name=name,
                    ops=self._ops,
                )
            )
        return ret

    @torch.no_grad()
    def get_logs(self, label: str) -> dict[str, float | torch.Tensor | Image]:
        logs = self._gen_agg.get_logs("")
        preds = self._get_target_gen_pairs()
        bias_map_key = "bias_map"
        rmse_all_channels = {}
        wandb = WandB.get_instance()
        for pred in preds:
            bias_data = pred.bias().cpu().numpy()
            vmin_bias, vmax_bias = get_cmap_limits(bias_data, diverging=True)
            bias_fig = plot_imshow(
                bias_data, vmin=vmin_bias, vmax=vmax_bias, cmap="RdBu_r"
            )
            bias_image = wandb.Image(
                bias_fig,
                caption=self._get_caption(
                    bias_map_key, pred.name, vmin_bias, vmax_bias
                ),
            )
            plt.close("all")
            rmse_all_channels[pred.name] = pred.rmse()
            logs.update({f"rmse/{pred.name}": rmse_all_channels[pred.name]})
            if self._target == "denorm":
                logs.update(
                    {
                        f"{bias_map_key}/{pred.name}": bias_image,
                        f"bias/{pred.name}": pred.weighted_mean_bias(),
                    }
                )
        if self._target == "norm":
            metric_name = "rmse/channel_mean"
            if self._channel_mean_names is None:
                values_to_average = list(rmse_all_channels.values())
            else:
                values_to_average = [
                    rmse_all_channels[name] for name in self._channel_mean_names
                ]
            logs.update({metric_name: sum(values_to_average) / len(values_to_average)})

        if len(label) != 0:
            return {f"{label}/{key}": logs[key] for key in logs}
        return logs

    def _get_caption(self, key: str, name: str, vmin: float, vmax: float) -> str:
        if name in self._variable_metadata:
            caption_name = self._variable_metadata[name].long_name
            units = self._variable_metadata[name].units
        else:
            caption_name, units = name, "unknown_units"
        caption = self._image_captions[key].format(name=caption_name, units=units)
        caption += f" vmin={vmin:.4g}, vmax={vmax:.4g}."
        return caption

    def get_dataset(self) -> xr.Dataset:
        data = {}
        preds = self._get_target_gen_pairs()
        for pred in preds:
            if pred.name in self._variable_metadata:
                long_name = self._variable_metadata[pred.name].long_name
                units = self._variable_metadata[pred.name].units
            else:
                long_name = pred.name
                units = "unknown_units"
            gen_metadata = VariableMetadata(long_name=long_name, units=units)._asdict()
            bias_metadata = self._variable_metadata.get(
                pred.name, VariableMetadata(long_name=long_name, units=units)
            )._asdict()
            data.update(
                {
                    f"bias_map-{pred.name}": xr.DataArray(
                        pred.bias().cpu(),
                        dims=self._horizontal_dims,
                        attrs=bias_metadata,
                    ),
                    f"gen_map-{pred.name}": xr.DataArray(
                        pred.gen.cpu(),
                        dims=self._horizontal_dims,
                        attrs=gen_metadata,
                    ),
                }
            )
        return xr.Dataset(data)



import torch

from fme.ace.stepper import TrainOutput
from fme.core.device import get_device
from fme.core.distributed import Distributed
from fme.core.generics.aggregator import AggregatorABC


class TrainAggregator(AggregatorABC[TrainOutput]):
    """
    Aggregates statistics for the first timestep.

    To use, call `record_batch` on the results of each batch, then call
    `get_logs` to get a dictionary of statistics when you're done.
    """

    def __init__(self):
        self._n_batches = 0
        self._loss = torch.tensor(0.0, device=get_device())

    @torch.no_grad()
    def record_batch(self, batch: TrainOutput):
        self._loss += batch.metrics["loss"]
        self._n_batches += 1

    @torch.no_grad()
    def get_logs(self, label: str) -> dict[str, torch.Tensor]:
        """
        Returns logs as can be reported to WandB.

        Args:
            label: Label to prepend to all log keys.
        """
        logs = {f"{label}/mean/loss": self._loss / self._n_batches}
        dist = Distributed.get_instance()
        for key in sorted(logs.keys()):
            logs[key] = float(dist.reduce_mean(logs[key].detach()).cpu().numpy())
        return logs

    @torch.no_grad()
    def flush_diagnostics(self, subdir: str | None) -> None:
        pass



import matplotlib.pyplot as plt
import pytest
import torch
import torch_harmonics

import fme
from fme.ace.aggregator.inference.spectrum import (
    PairedSphericalPowerSpectrumAggregator,
    SphericalPowerSpectrumAggregator,
    get_positive_and_negative_power_bias,
    get_smallest_scale_power_bias,
)
from fme.core.gridded_ops import LatLonOperations
from fme.core.metrics import spherical_power_spectrum

DEVICE = fme.get_device()


def get_gridded_operations(nlat: int, nlon: int):
    return LatLonOperations(torch.ones(nlat, nlon))


def test_spherical_power_spectrum_aggregator():
    nlat = 8
    nlon = 16
    grid = "legendre-gauss"
    gridded_operations = get_gridded_operations(nlat, nlon)
    agg = SphericalPowerSpectrumAggregator(gridded_operations)
    data = {"a": torch.randn(2, 2, nlat, nlon, device=fme.get_device())}
    data2 = {"a": torch.randn(2, 3, nlat, nlon, device=fme.get_device())}
    agg.record_batch(data)
    agg.record_batch(data2)
    result = agg.get_mean()
    assert "a" in result
    assert result["a"].shape == (nlat,)

    sht = torch_harmonics.RealSHT(nlat, nlon, grid=grid).to(fme.get_device())
    data_concat = torch.cat([data["a"], data2["a"]], dim=1)
    expected_value = torch.mean(spherical_power_spectrum(data_concat, sht), dim=(0, 1))
    torch.testing.assert_close(result["a"], expected_value)


@pytest.mark.parametrize("report_plot", [True, False])
def test_paired_spherical_power_spectrum_aggregator(report_plot: bool):
    nlat = 8
    nlon = 16
    gridded_operations = get_gridded_operations(nlat, nlon)
    agg = PairedSphericalPowerSpectrumAggregator(
        gridded_operations, report_plot=report_plot
    )
    data = {"a": torch.randn(2, 3, nlat, nlon, device=fme.get_device())}
    agg.record_batch(data, data, None, None)
    result = agg.get_logs("spectrum")
    if report_plot:
        assert isinstance(result["spectrum/a"], plt.Figure)
    else:
        assert "spectrum/a" not in result


@pytest.mark.parametrize(
    "gen_spectrum, target_spectrum, expected_positive_bias, expected_negative_bias",
    [
        pytest.param(
            torch.tensor([1.0, 1.2, 1.4], device=DEVICE),
            torch.tensor([1, 1, 1], device=DEVICE),
            0.2,
            0.0,
            id="positive_bias",
        ),
        pytest.param(
            torch.tensor([1.0, 0.8, 0.6], device=DEVICE),
            torch.tensor([1, 1, 1], device=DEVICE),
            0.0,
            -0.2,
            id="negative_bias",
        ),
        pytest.param(
            torch.tensor([1.6, 0.8, 0.6], device=DEVICE),
            torch.tensor([1, 1, 1], device=DEVICE),
            0.2,
            -0.2,
            id="both_bias",
        ),
        pytest.param(
            torch.tensor([2.0, 2.4, 2.8], device=DEVICE),
            torch.tensor([2, 2, 2], device=DEVICE),
            0.2,
            0.0,
            id="positive_bias_with_scale",
        ),
    ],
)
def test_get_positive_and_negative_power_bias(
    gen_spectrum: torch.Tensor,
    target_spectrum: torch.Tensor,
    expected_positive_bias: float,
    expected_negative_bias: float,
):
    positive_bias, negative_bias = get_positive_and_negative_power_bias(
        gen_spectrum, target_spectrum
    )
    torch.testing.assert_close(positive_bias, expected_positive_bias)
    torch.testing.assert_close(negative_bias, expected_negative_bias)


@pytest.mark.parametrize(
    "gen_spectrum, target_spectrum, expected_bias",
    [
        pytest.param(
            torch.tensor([1.0, 1.2, 1.4], device=DEVICE),
            torch.tensor([1, 1, 1], device=DEVICE),
            0.4,
            id="positive_bias",
        ),
        pytest.param(
            torch.tensor([1.0, 0.8, 1.3], device=DEVICE),
            torch.tensor([1, 1, 1.3], device=DEVICE),
            0.0,
            id="no_bias",
        ),
    ],
)
def test_get_smallest_scale_power_bias(
    gen_spectrum: torch.Tensor,
    target_spectrum: torch.Tensor,
    expected_bias: float,
):
    bias = get_smallest_scale_power_bias(gen_spectrum, target_spectrum)
    torch.testing.assert_close(bias, expected_bias)



from datetime import timedelta

import numpy as np
import pandas as pd
import pytest

from fme.core.dataset.config import RepeatedInterval


def test_repeated_interval_int():
    interval = RepeatedInterval(interval_length=3, block_length=6, start=0)
    mask = interval.get_boolean_mask(length=18)
    expected_mask = np.array([True, True, True, False, False, False] * 3)
    np.testing.assert_array_equal(mask, expected_mask)


def test_repeated_interval_str():
    interval = RepeatedInterval(interval_length="1d", block_length="7d", start="2d")
    mask = interval.get_boolean_mask(length=21, timestep=timedelta(days=1))
    expected_mask = np.array([False, False, True, False, False, False, False] * 3)
    np.testing.assert_array_equal(mask, expected_mask)


def test_repeated_interval_mixed_types():
    with pytest.raises(ValueError):
        RepeatedInterval(interval_length=3, block_length="6d", start=0)


@pytest.mark.parametrize("interval, block, start", [(4, 6, 3), ("2d", "3d", "2d")])
def test_repeated_interval_invalid_interval_start(interval, block, start):
    """start + interval exceeds length of block"""
    interval = RepeatedInterval(
        interval_length=interval, block_length=block, start=start
    )
    with pytest.raises(ValueError):
        interval.get_boolean_mask(length=18, timestep=timedelta(days=1))


def test_repeated_interval_zero_length():
    interval = RepeatedInterval(interval_length=0, block_length=6, start=0)
    mask = interval.get_boolean_mask(length=18)
    expected_mask = np.array([False] * 18)
    np.testing.assert_array_equal(mask, expected_mask)


def test_repeated_interval_partial_block():
    interval = RepeatedInterval(interval_length=3, block_length=6, start=0)
    mask = interval.get_boolean_mask(length=20)
    expected_mask = np.array([True, True, True, False, False, False] * 3 + [True, True])
    np.testing.assert_array_equal(mask, expected_mask)


def test_repeated_interval_no_timestep_fails_for_timedelta_lengths():
    interval = RepeatedInterval(interval_length="1d", block_length="7d", start="0d")
    with pytest.raises(ValueError):
        interval.get_boolean_mask(length=20)


@pytest.mark.parametrize("timestep", ["2h", "150m", "5h", "12h"])
def test_invalid_timesteps(timestep):
    """
    Test that timesteps that don't evenly divide into some or all
    arguments raise a ValueError
    """
    timestep = pd.to_timedelta(timestep)
    with pytest.raises(ValueError):
        RepeatedInterval(
            interval_length="5h", start="4h", block_length="10h"
        ).get_boolean_mask(length=18, timestep=timestep)



import dataclasses
import datetime
import logging
import warnings
from collections.abc import Callable, Mapping, Sequence
from typing import Protocol

import torch
import xarray as xr

from fme.ace.data_loading.batch_data import PairedData, PrognosticState
from fme.core.coordinates import LatLonCoordinates
from fme.core.dataset_info import DatasetInfo
from fme.core.diagnostics import get_reduced_diagnostics, write_reduced_diagnostics
from fme.core.generics.aggregator import (
    InferenceAggregatorABC,
    InferenceLog,
    InferenceLogs,
)
from fme.core.gridded_ops import LatLonOperations
from fme.core.typing_ import TensorDict, TensorMapping
from fme.core.wandb import Table, WandB

from ..one_step.reduced import MeanAggregator as OneStepMeanAggregator
from .annual import GlobalMeanAnnualAggregator, PairedGlobalMeanAnnualAggregator
from .enso import (
    EnsoCoefficientEvaluatorAggregator,
    LatLonRegion,
    PairedRegionalIndexAggregator,
    RegionalIndexAggregator,
)
from .histogram import HistogramAggregator
from .reduced import MeanAggregator, SingleTargetMeanAggregator
from .seasonal import SeasonalAggregator
from .spectrum import PairedSphericalPowerSpectrumAggregator
from .time_mean import TimeMeanAggregator, TimeMeanEvaluatorAggregator
from .video import VideoAggregator
from .zonal_mean import ZonalMeanAggregator

wandb = WandB.get_instance()
APPROXIMATELY_TWO_YEARS = datetime.timedelta(days=730)
SLIGHTLY_LESS_THAN_FIVE_YEARS = datetime.timedelta(days=1800)
NINO34_LAT = (-5, 5)
NINO34_LON = (190, 240)


class _Aggregator(Protocol):
    @torch.no_grad()
    def record_batch(
        self,
        data: TensorMapping,
    ): ...

    @torch.no_grad()
    def get_logs(self, label: str): ...

    @torch.no_grad()
    def get_dataset(self) -> xr.Dataset: ...


class _EvaluatorAggregator(Protocol):
    @torch.no_grad()
    def record_batch(
        self,
        target_data: TensorMapping,
        gen_data: TensorMapping,
        target_data_norm: TensorMapping,
        gen_data_norm: TensorMapping,
        i_time_start: int = 0,
    ): ...

    @torch.no_grad()
    def get_logs(self, label: str): ...

    @torch.no_grad()
    def get_dataset(self) -> xr.Dataset: ...


class _TimeDependentAggregator(Protocol):
    @torch.no_grad()
    def record_batch(
        self,
        time: xr.DataArray,
        data: TensorMapping,
    ): ...

    @torch.no_grad()
    def get_logs(self, label: str): ...

    @torch.no_grad()
    def get_dataset(self) -> xr.Dataset: ...


class _TimeDependentEvaluatorAggregator(Protocol):
    @torch.no_grad()
    def record_batch(
        self,
        time: xr.DataArray,
        target_data: TensorMapping,
        gen_data: TensorMapping,
    ): ...

    @torch.no_grad()
    def get_logs(self, label: str): ...

    @torch.no_grad()
    def get_dataset(self) -> xr.Dataset: ...


@dataclasses.dataclass
class InferenceEvaluatorAggregatorConfig:
    """
    Configuration for inference evaluator aggregator.

    Parameters:
        log_histograms: Whether to log histograms of the targets and predictions.
        log_video: Whether to log videos of the state evolution.
        log_extended_video: Whether to log wandb videos of the predictions with
            statistical metrics, only done if log_video is True.
        log_zonal_mean_images: Whether to log zonal-mean images (hovmollers) with a
            time dimension.
        log_seasonal_means: Whether to log seasonal mean metrics and images.
        log_global_mean_time_series: Whether to log global mean time series metrics.
        log_global_mean_norm_time_series: Whether to log the normalized global mean
            time series metrics.
        monthly_reference_data: Path to monthly reference data to compare against.
        time_mean_reference_data: Path to reference time means to compare against.
    """

    log_histograms: bool = False
    log_video: bool = False
    log_extended_video: bool = False
    log_zonal_mean_images: bool = True
    log_seasonal_means: bool = False
    log_global_mean_time_series: bool = True
    log_global_mean_norm_time_series: bool = True
    monthly_reference_data: str | None = None
    time_mean_reference_data: str | None = None

    def build(
        self,
        dataset_info: DatasetInfo,
        n_timesteps: int,
        initial_time: xr.DataArray,
        normalize: Callable[[TensorMapping], TensorDict],
        output_dir: str | None = None,
        record_step_20: bool = False,
        channel_mean_names: Sequence[str] | None = None,
        save_diagnostics: bool = True,
    ) -> "InferenceEvaluatorAggregator":
        if save_diagnostics and output_dir is None:
            raise ValueError("Output directory must be set to save diagnostics.")
        if self.monthly_reference_data is None:
            monthly_reference_data = None
        else:
            monthly_reference_data = xr.open_dataset(
                self.monthly_reference_data, decode_timedelta=False
            )
        if self.time_mean_reference_data is None:
            time_mean = None
        else:
            time_mean = xr.open_dataset(
                self.time_mean_reference_data, decode_timedelta=False
            )

        if n_timesteps > 2**15 and self.log_zonal_mean_images:
            # matplotlib raises an error if image size is too large, and we plot
            # one pixel per timestep in the zonal mean images.
            warnings.warn(
                "Disabling zonal mean images logging due to large number of timesteps"
                f" (n_timesteps={n_timesteps}). Set log_zonal_mean_images=False or "
                "decrease n_timesteps to below 2**15 to avoid this warning."
            )
            log_zonal_mean_images = False
        else:
            log_zonal_mean_images = self.log_zonal_mean_images

        return InferenceEvaluatorAggregator(
            dataset_info=dataset_info,
            n_timesteps=n_timesteps,
            initial_time=initial_time,
            output_dir=output_dir,
            log_histograms=self.log_histograms,
            log_video=self.log_video,
            enable_extended_videos=self.log_extended_video,
            log_zonal_mean_images=log_zonal_mean_images,
            log_seasonal_means=self.log_seasonal_means,
            log_global_mean_time_series=self.log_global_mean_time_series,
            log_global_mean_norm_time_series=self.log_global_mean_norm_time_series,
            monthly_reference_data=monthly_reference_data,
            time_mean_reference_data=time_mean,
            record_step_20=record_step_20,
            channel_mean_names=channel_mean_names,
            normalize=normalize,
            save_diagnostics=save_diagnostics,
        )


class InferenceEvaluatorAggregator(
    InferenceAggregatorABC[PairedData | PrognosticState, PairedData]
):
    """
    Aggregates statistics for inference comparing a generated and target series.

    To use, call `record_batch` on the results of each batch, then call
    `get_logs` to get a dictionary of statistics when you're done.
    """

    def __init__(
        self,
        dataset_info: DatasetInfo,
        n_timesteps: int,
        initial_time: xr.DataArray,
        normalize: Callable[[TensorMapping], TensorDict],
        output_dir: str | None = None,
        record_step_20: bool = False,
        log_video: bool = False,
        enable_extended_videos: bool = False,
        log_zonal_mean_images: bool = False,
        log_seasonal_means: bool = False,
        log_global_mean_time_series: bool = True,
        log_global_mean_norm_time_series: bool = True,
        monthly_reference_data: xr.Dataset | None = None,
        log_histograms: bool = False,
        time_mean_reference_data: xr.Dataset | None = None,
        channel_mean_names: Sequence[str] | None = None,
        log_nino34_index: bool = True,
        save_diagnostics: bool = True,
    ):
        """
        Args:
            dataset_info: Dataset coordinates and metadata.
            n_timesteps: Number of timesteps of inference that will be run.
            initial_time: Initial time for each sample.
            output_dir: Directory to save diagnostic output.
            normalize: Normalization function to use.
            record_step_20: Whether to record the mean of the 20th steps.
            log_video: Whether to log videos of the state evolution.
            enable_extended_videos: Whether to log videos of statistical
                metrics of state evolution
            log_zonal_mean_images: Whether to log zonal-mean images (hovmollers) with a
                time dimension.
            log_seasonal_means: Whether to log seasonal means metrics and images.
            log_global_mean_time_series: Whether to log global mean time series metrics.
            log_global_mean_norm_time_series: Whether to log the normalized global mean
                time series metrics.
            monthly_reference_data: Reference monthly data for computing target stats.
            log_histograms: Whether to aggregate histograms.
            data_grid: The grid type of the data, used for spherical power spectrum.
            time_mean_reference_data: Reference time means for computing bias stats.
            channel_mean_names: Names over which to compute channel means. If not
                provided, all available variables will be used.
            log_nino34_index: Whether to log the Nino34 index.
            save_diagnostics: Whether to save reduced diagnostics to disk.
        """
        if save_diagnostics and output_dir is None:
            raise ValueError("Output directory must be set to save diagnostics")
        self._channel_mean_names = channel_mean_names
        self._aggregators: dict[str, _EvaluatorAggregator] = {}
        self._time_dependent_aggregators: dict[
            str, _TimeDependentEvaluatorAggregator
        ] = {}
        self._save_diagnostics = save_diagnostics
        self._output_dir = output_dir
        timestep = dataset_info.timestep
        horizontal_coordinates = dataset_info.horizontal_coordinates
        self._coords = horizontal_coordinates.coords
        ops = dataset_info.gridded_operations
        self._log_time_series = (
            log_global_mean_time_series or log_global_mean_norm_time_series
        )
        if log_global_mean_time_series:
            self._aggregators["mean"] = MeanAggregator(
                ops,
                target="denorm",
                n_timesteps=n_timesteps,
                variable_metadata=dataset_info.variable_metadata,
            )
        if log_global_mean_norm_time_series:
            self._aggregators["mean_norm"] = MeanAggregator(
                ops,
                target="norm",
                n_timesteps=n_timesteps,
                variable_metadata=dataset_info.variable_metadata,
            )
        if record_step_20:
            self._aggregators["mean_step_20"] = OneStepMeanAggregator(
                ops, target_time=20
            )
        try:
            self._aggregators["power_spectrum"] = (
                PairedSphericalPowerSpectrumAggregator(
                    gridded_operations=ops,
                    report_plot=True,
                )
            )
        except NotImplementedError:
            logging.warning(
                "Power spectrum aggregator not implemented for this grid type, "
                "omitting."
            )
        if log_zonal_mean_images:
            if ops.zonal_mean is None:
                logging.warning(
                    "Zonal mean aggregator not implemented for this grid type, "
                    "omitting."
                )
            else:
                self._aggregators["zonal_mean"] = ZonalMeanAggregator(
                    zonal_mean=ops.zonal_mean,
                    n_timesteps=n_timesteps,
                    variable_metadata=dataset_info.variable_metadata,
                )
        if isinstance(horizontal_coordinates, LatLonCoordinates):
            if log_video:
                self._aggregators["video"] = VideoAggregator(
                    n_timesteps=n_timesteps,
                    enable_extended_videos=enable_extended_videos,
                    variable_metadata=dataset_info.variable_metadata,
                )
        self._aggregators["time_mean"] = TimeMeanEvaluatorAggregator(
            ops,
            horizontal_dims=horizontal_coordinates.dims,
            variable_metadata=dataset_info.variable_metadata,
            reference_means=time_mean_reference_data,
        )
        self._aggregators["time_mean_norm"] = TimeMeanEvaluatorAggregator(
            ops,
            horizontal_dims=horizontal_coordinates.dims,
            target="norm",
            variable_metadata=dataset_info.variable_metadata,
        )
        if log_histograms:
            self._aggregators["histogram"] = HistogramAggregator()
        if log_seasonal_means:
            self._time_dependent_aggregators["seasonal"] = SeasonalAggregator(
                ops=ops,
                variable_metadata=dataset_info.variable_metadata,
            )
        if n_timesteps * timestep > APPROXIMATELY_TWO_YEARS:
            self._time_dependent_aggregators["annual"] = (
                PairedGlobalMeanAnnualAggregator(
                    ops=ops,
                    timestep=timestep,
                    variable_metadata=dataset_info.variable_metadata,
                    monthly_reference_data=monthly_reference_data,
                )
            )
            if (
                isinstance(horizontal_coordinates, LatLonCoordinates)
                and isinstance(ops, LatLonOperations)
                and log_nino34_index
            ):
                nino34_region = LatLonRegion(
                    lat_bounds=NINO34_LAT,
                    lon_bounds=NINO34_LON,
                    lat=horizontal_coordinates.lat,
                    lon=horizontal_coordinates.lon,
                )
                self._time_dependent_aggregators["enso_index"] = (
                    PairedRegionalIndexAggregator(
                        target_aggregator=RegionalIndexAggregator(
                            regional_weights=nino34_region.regional_weights,
                            regional_mean=ops.regional_area_weighted_mean,
                        ),
                        prediction_aggregator=RegionalIndexAggregator(
                            regional_weights=nino34_region.regional_weights,
                            regional_mean=ops.regional_area_weighted_mean,
                        ),
                    )
                )
        if n_timesteps * timestep > SLIGHTLY_LESS_THAN_FIVE_YEARS:
            self._time_dependent_aggregators["enso_coefficient"] = (
                EnsoCoefficientEvaluatorAggregator(
                    initial_time,
                    n_timesteps - 1,
                    timestep,
                    gridded_operations=ops,
                    variable_metadata=dataset_info.variable_metadata,
                )
            )
        self._summary_aggregators = {
            name: agg
            for name, agg in list(self._aggregators.items())
            + list(self._time_dependent_aggregators.items())
            if name not in ["mean", "mean_norm"]
        }
        self._n_timesteps_seen = 0
        self._normalize = normalize

    @property
    def log_time_series(self) -> bool:
        return self._log_time_series

    @torch.no_grad()
    def record_batch(
        self,
        data: PairedData,
    ) -> InferenceLogs:
        if len(data.prediction) == 0:
            raise ValueError("No prediction values in data")
        if len(data.target) == 0:
            raise ValueError("No target values in data")
        target_data = data.target
        target_data_norm = self._normalize(target_data)
        gen_data_norm = self._normalize(data.prediction)
        for aggregator in self._aggregators.values():
            aggregator.record_batch(
                target_data=target_data,
                gen_data=data.prediction,
                target_data_norm=target_data_norm,
                gen_data_norm=gen_data_norm,
                i_time_start=self._n_timesteps_seen,
            )
        for time_dependent_aggregator in self._time_dependent_aggregators.values():
            time_dependent_aggregator.record_batch(
                time=data.time,
                target_data=target_data,
                gen_data=data.prediction,
            )
        n_times = data.time.shape[1]
        logs = self._get_inference_logs_slice(
            step_slice=slice(self._n_timesteps_seen, self._n_timesteps_seen + n_times),
        )
        self._n_timesteps_seen += n_times
        return logs

    def record_initial_condition(
        self,
        initial_condition: PairedData | PrognosticState,
    ) -> InferenceLogs:
        if self._n_timesteps_seen != 0:
            raise RuntimeError(
                "record_initial_condition may only be called once, "
                "before recording any batches"
            )
        if isinstance(initial_condition, PairedData):
            target_data = initial_condition.target
            target_data_norm = self._normalize(target_data)
            gen_data = initial_condition.prediction
            gen_data_norm = self._normalize(gen_data)
            n_times = initial_condition.time.shape[1]
        else:
            batch_data = initial_condition.as_batch_data()
            target_data = batch_data.data
            target_data_norm = self._normalize(target_data)
            gen_data = target_data
            gen_data_norm = target_data_norm
            n_times = batch_data.time.shape[1]
        for aggregator_name in ["mean", "mean_norm"]:
            aggregator = self._aggregators.get(aggregator_name)
            if aggregator is not None:
                aggregator.record_batch(
                    target_data=target_data,
                    gen_data=gen_data,
                    target_data_norm=target_data_norm,
                    gen_data_norm=gen_data_norm,
                    i_time_start=0,
                )
        logs = self._get_inference_logs_slice(
            step_slice=slice(self._n_timesteps_seen, self._n_timesteps_seen + n_times),
        )
        self._n_timesteps_seen = n_times
        return logs

    def get_summary_logs(self) -> InferenceLog:
        logs = {}
        for name, aggregator in self._summary_aggregators.items():
            logging.info(f"Getting summary logs for {name} aggregator")
            logs.update(aggregator.get_logs(label=name))
        return logs

    @torch.no_grad()
    def _get_logs(self):
        """
        Returns logs as can be reported to WandB.
        """
        logs = {}
        for name, aggregator in self._aggregators.items():
            logs.update(aggregator.get_logs(label=name))
        for name, time_dependent_aggregator in self._time_dependent_aggregators.items():
            logs.update(time_dependent_aggregator.get_logs(label=name))
        return logs

    @torch.no_grad()
    def _get_inference_logs_slice(self, step_slice: slice):
        """
        Returns a subset of the time series for applicable metrics
        for a specific slice of as can be reported to WandB.

        Args:
            step_slice: Timestep slice to determine the time series subset.

        Returns:
            Tuple of start index and list of logs.
        """
        logs = {}
        for name, aggregator in self._aggregators.items():
            if isinstance(aggregator, MeanAggregator):
                logs.update(aggregator.get_logs(label=name, step_slice=step_slice))
        return to_inference_logs(logs)

    @torch.no_grad()
    def flush_diagnostics(self, subdir: str | None = None):
        if self._save_diagnostics:
            reduced_diagnostics = get_reduced_diagnostics(
                sub_aggregators=(self._aggregators | self._time_dependent_aggregators),
                coords=self._coords,
            )
            if self._output_dir is not None:
                write_reduced_diagnostics(
                    reduced_diagnostics=reduced_diagnostics,
                    output_dir=self._output_dir,
                    subdir=subdir,
                )
            else:
                raise ValueError("Output directory not set.")


def to_inference_logs(
    log: Mapping[str, Table | float | int],
) -> list[dict[str, float | int]]:
    # we have a dictionary which contains WandB tables
    # which we will convert to a list of dictionaries, one for each
    # row in the tables. Any scalar values will be reported in the last
    # dictionary.
    n_rows = 0
    for val in log.values():
        if isinstance(val, Table):
            n_rows = max(n_rows, len(val.data))
    logs: list[dict[str, float | int]] = []
    for i in range(max(1, n_rows)):  # need at least one for non-series values
        logs.append({})
    for key, val in log.items():
        if isinstance(val, Table):
            for i, row in enumerate(val.data):
                for j, col in enumerate(val.columns):
                    key_without_table_name = key[: key.rfind("/")]
                    logs[i][f"{key_without_table_name}/{col}"] = row[j]
        else:
            logs[-1][key] = val
    return logs


def table_to_logs(table: Table) -> list[dict[str, float | int]]:
    """
    Converts a WandB table into a list of dictionaries.
    """
    logs = []
    for row in table.data:
        logs.append({table.columns[i]: row[i] for i in range(len(row))})
    return logs


@dataclasses.dataclass
class InferenceAggregatorConfig:
    """
    Configuration for inference aggregator.

    Parameters:
        time_mean_reference_data: Path to reference time means to compare against.
        log_global_mean_time_series: Whether to log global mean time series metrics.
    """

    time_mean_reference_data: str | None = None
    log_global_mean_time_series: bool = True

    def build(
        self,
        dataset_info: DatasetInfo,
        n_timesteps: int,
        output_dir: str,
    ) -> "InferenceAggregator":
        if self.time_mean_reference_data is not None:
            time_means = xr.open_dataset(
                self.time_mean_reference_data,
                decode_timedelta=False,
            )
        else:
            time_means = None
        return InferenceAggregator(
            dataset_info=dataset_info,
            n_timesteps=n_timesteps,
            output_dir=output_dir,
            time_mean_reference_data=time_means,
            log_global_mean_time_series=self.log_global_mean_time_series,
        )


class InferenceAggregator(
    InferenceAggregatorABC[
        PrognosticState,
        PairedData,
    ]
):
    """
    Aggregates statistics on a single timeseries of data.

    To use, call `record_batch` on the results of each batch, then call
    `get_logs` to get a dictionary of statistics when you're done.
    """

    def __init__(
        self,
        dataset_info: DatasetInfo,
        n_timesteps: int,
        save_diagnostics: bool = True,
        output_dir: str | None = None,
        time_mean_reference_data: xr.Dataset | None = None,
        log_global_mean_time_series: bool = True,
    ):
        """
        Args:
            dataset_info: The coordinates of the dataset.
            n_timesteps: Number of timesteps in the model.
            save_diagnostics: Whether to save diagnostics.
            output_dir: Directory to save diagnostic output.
            time_mean_reference_data: Reference time means for computing bias stats.
            log_global_mean_time_series: Whether to log global mean time series metrics.
        """
        if save_diagnostics and output_dir is None:
            raise ValueError("Output directory must be set to save diagnostics")
        self._log_time_series = log_global_mean_time_series
        horizontal_coordinates = dataset_info.horizontal_coordinates
        self._coords = horizontal_coordinates.coords
        self._save_diagnostics = save_diagnostics
        self._output_dir = output_dir
        aggregators: dict[str, _Aggregator] = {}
        gridded_operations = dataset_info.gridded_operations
        if log_global_mean_time_series:
            aggregators["mean"] = SingleTargetMeanAggregator(
                gridded_operations,
                n_timesteps=n_timesteps,
            )
        aggregators["time_mean"] = TimeMeanAggregator(
            gridded_operations=gridded_operations,
            variable_metadata=dataset_info.variable_metadata,
            reference_means=time_mean_reference_data,
        )
        aggregators["annual"] = GlobalMeanAnnualAggregator(
            gridded_operations,
            dataset_info.timestep,
            dataset_info.variable_metadata,
        )
        if (
            isinstance(horizontal_coordinates, LatLonCoordinates)
            and isinstance(gridded_operations, LatLonOperations)
            and n_timesteps * dataset_info.timestep > APPROXIMATELY_TWO_YEARS
        ):
            nino34_region = LatLonRegion(
                lat_bounds=NINO34_LAT,
                lon_bounds=NINO34_LON,
                lat=horizontal_coordinates.lat,
                lon=horizontal_coordinates.lon,
            )
            aggregators["enso_index"] = RegionalIndexAggregator(
                regional_weights=nino34_region.regional_weights,
                regional_mean=gridded_operations.regional_area_weighted_mean,
            )
        self._aggregators = aggregators
        self._summary_aggregators = {
            name: aggregators[name]
            for name in ["time_mean", "annual", "enso_index"]
            if name in aggregators
        }
        self._time_dependent_aggregator_names = ["annual", "enso_index"]
        self._n_timesteps_seen = 0

    @property
    def log_time_series(self) -> bool:
        return self._log_time_series

    @torch.no_grad()
    def record_batch(self, data: PairedData) -> InferenceLogs:
        """
        Record a batch of data.

        Args:
            data: Batch of data to record.
        """
        if len(data.prediction) == 0:
            raise ValueError("data is empty")
        for name in self._aggregators:
            if name in self._time_dependent_aggregator_names:
                self._aggregators[name].record_batch(
                    time=data.time, data=data.prediction
                )
            else:
                self._aggregators[name].record_batch(
                    data=data.prediction,
                    i_time_start=self._n_timesteps_seen,
                )
        n_times = data.time.shape[1]
        logs = self._get_inference_logs_slice(
            step_slice=slice(self._n_timesteps_seen, self._n_timesteps_seen + n_times),
        )
        self._n_timesteps_seen += n_times
        return logs

    def record_initial_condition(
        self,
        initial_condition: PrognosticState,
    ) -> InferenceLogs:
        if self._n_timesteps_seen != 0:
            raise RuntimeError(
                "record_initial_condition may only be called once, "
                "before recording any batches"
            )
        batch_data = initial_condition.as_batch_data()
        if "mean" in self._aggregators:
            self._aggregators["mean"].record_batch(
                data=batch_data.data,
                i_time_start=0,
            )
        n_times = batch_data.time.shape[1]
        logs = self._get_inference_logs_slice(
            step_slice=slice(self._n_timesteps_seen, self._n_timesteps_seen + n_times),
        )
        self._n_timesteps_seen = n_times
        return logs

    def get_summary_logs(self) -> InferenceLog:
        logs = {}
        for name, aggregator in self._summary_aggregators.items():
            logging.info(f"Getting summary logs for {name} aggregator")
            logs.update(aggregator.get_logs(label=name))
        return logs

    @torch.no_grad()
    def _get_logs(self):
        """
        Returns logs as can be reported to WandB.
        """
        logs = {}
        for name, aggregator in self._aggregators.items():
            logs.update(aggregator.get_logs(label=name))
        return logs

    @torch.no_grad()
    def _get_inference_logs(self) -> list[dict[str, float | int]]:
        """
        Returns a list of logs to report to WandB.

        This is done because in inference, we use the wandb step
        as the time step, meaning we need to re-organize the logged data
        from tables into a list of dictionaries.
        """
        return to_inference_logs(self._get_logs())

    @torch.no_grad()
    def _get_inference_logs_slice(self, step_slice: slice):
        """
        Returns a subset of the time series for applicable metrics
        for a specific slice of as can be reported to WandB.

        Args:
            step_slice: Timestep slice to determine the time series subset.
        """
        logs = {}
        for name, aggregator in self._aggregators.items():
            if isinstance(aggregator, SingleTargetMeanAggregator):
                logs.update(aggregator.get_logs(label=name, step_slice=step_slice))
        return to_inference_logs(logs)

    @torch.no_grad()
    def flush_diagnostics(self, subdir: str | None = None):
        if self._save_diagnostics:
            reduced_diagnostics = get_reduced_diagnostics(
                sub_aggregators=self._aggregators,
                coords=self._coords,
            )
            if self._output_dir is not None:
                write_reduced_diagnostics(
                    reduced_diagnostics=reduced_diagnostics,
                    output_dir=self._output_dir,
                    subdir=subdir,
                )
            else:
                raise ValueError("Output directory is not set.")



import logging
import warnings
from collections import defaultdict

import matplotlib.pyplot as plt
import torch
import xarray as xr

from fme.core.distributed import Distributed
from fme.core.gridded_ops import GriddedOperations
from fme.core.metrics import spherical_power_spectrum
from fme.core.typing_ import TensorMapping


class SphericalPowerSpectrumAggregator:
    """Average the power spectrum over batch and time dimensions."""

    def __init__(self, gridded_operations: GriddedOperations):
        self._real_sht = gridded_operations.get_real_sht()
        self._power_spectrum: dict[str, torch.Tensor] = {}
        self._counts: dict[str, int] = defaultdict(int)

    @torch.no_grad()
    def record_batch(self, data: TensorMapping):
        for name in data:
            batch_size = data[name].shape[0]
            time_size = data[name].shape[1]
            power_spectrum = spherical_power_spectrum(data[name], self._real_sht)
            mean_power_spectrum = torch.mean(power_spectrum, dim=(0, 1))
            new_count = batch_size * time_size
            if name not in self._power_spectrum:
                self._power_spectrum[name] = mean_power_spectrum
            else:
                weighted_average = (
                    new_count * mean_power_spectrum
                    + self._counts[name] * self._power_spectrum[name]
                ) / (new_count + self._counts[name])
                self._power_spectrum[name] = weighted_average
            self._counts[name] += new_count

    def get_mean(self) -> dict[str, torch.Tensor]:
        dist = Distributed.get_instance()
        logs = {}
        sorted_names = sorted(list(self._power_spectrum))
        for name in sorted_names:
            _mean_spectrum = self._power_spectrum[name]
            if dist.world_size > 1:
                # assuming same count on all workers
                _mean_spectrum = dist.reduce_mean(_mean_spectrum)
            logs[name] = _mean_spectrum
        return logs


class PairedSphericalPowerSpectrumAggregator:
    """Record batches and return plots for paired prediction and target data."""

    def __init__(
        self,
        gridded_operations: GriddedOperations,
        report_plot: bool,
    ):
        self._gen_aggregator = SphericalPowerSpectrumAggregator(gridded_operations)
        self._target_aggregator = SphericalPowerSpectrumAggregator(gridded_operations)
        self._report_plot = report_plot

    @torch.no_grad()
    def record_batch(
        self,
        target_data: TensorMapping,
        gen_data: TensorMapping,
        target_data_norm: TensorMapping,
        gen_data_norm: TensorMapping,
        i_time_start: int = 0,
    ):
        self._gen_aggregator.record_batch(gen_data)
        self._target_aggregator.record_batch(target_data)

    @torch.no_grad()
    def get_logs(self, label: str) -> dict[str, plt.Figure]:
        logs = {}
        gen_spectrum = self._gen_aggregator.get_mean()
        target_spectrum = self._target_aggregator.get_mean()
        if self._report_plot:
            for name in gen_spectrum:
                gen_spectrum_cpu = gen_spectrum[name].cpu()
                if name not in target_spectrum:
                    warnings.warn(f"Missing power spectrum target data for {name}")
                    target_spectrum_cpu = None
                else:
                    target_spectrum_cpu = target_spectrum[name].cpu()
                fig = _plot_spectrum_pair(gen_spectrum_cpu, target_spectrum_cpu)
                logs[f"{label}/{name}"] = fig
                plt.close(fig)
        metrics = _get_spectrum_metrics(gen_spectrum, target_spectrum)
        for name, value in metrics.items():
            logs[f"{label}/{name}"] = value
        return logs

    @torch.no_grad()
    def get_dataset(self) -> xr.Dataset:
        logging.debug(
            "get_dataset not implemented for PairedSphericalPowerSpectrumAggregator. "
            "Returning an empty dataset."
        )
        return xr.Dataset()


def _get_spectrum_metrics(
    gen_spectrum: dict[str, torch.Tensor],
    target_spectrum: dict[str, torch.Tensor],
) -> dict[str, float]:
    """
    Compute metrics for the spectrum.

    Args:
        gen_spectrum: Dictionary of 1-dimensional generated mean power spectra.
        target_spectrum: Dictionary of 1-dimensional target mean power spectra.

    Returns:
        Dictionary of metrics.
    """
    metrics = {}
    for name in gen_spectrum:
        if len(gen_spectrum[name].shape) != 1:
            raise ValueError(
                f"Expected 1-dimensional power spectrum for {name}, "
                f"got {gen_spectrum[name].shape}"
            )
        metrics[f"smallest_scale_norm_bias/{name}"] = get_smallest_scale_power_bias(
            gen_spectrum[name], target_spectrum[name]
        )
        positive_bias, negative_bias = get_positive_and_negative_power_bias(
            gen_spectrum[name], target_spectrum[name]
        )
        metrics[f"positive_norm_bias/{name}"] = positive_bias
        metrics[f"negative_norm_bias/{name}"] = negative_bias
        metrics[f"mean_abs_norm_bias/{name}"] = abs(positive_bias) + abs(negative_bias)
    return metrics


def get_smallest_scale_power_bias(
    gen_spectrum: torch.Tensor,
    target_spectrum: torch.Tensor,
) -> float:
    return float((gen_spectrum[-1] / target_spectrum[-1] - 1).mean().cpu())


def get_positive_and_negative_power_bias(
    gen_spectrum: torch.Tensor,
    target_spectrum: torch.Tensor,
) -> tuple[float, float]:
    """
    Compute the positive and negative power bias for the spectrum,
    normalized by the target spectrum.
    """
    ratio = gen_spectrum / target_spectrum - 1
    positive_bias = ratio[ratio > 0].sum() / target_spectrum.shape[0]
    negative_bias = ratio[ratio < 0].sum() / target_spectrum.shape[0]
    return float(positive_bias.cpu()), float(negative_bias.cpu())


def _plot_spectrum_pair(
    prediction: torch.Tensor, target: torch.Tensor | None
) -> plt.Figure:
    fig, ax = plt.subplots()
    ax.plot(prediction, "--", label="prediction", color="C1")
    if target is not None:
        ax.plot(target, "-", label="target", color="C0")
    ax.set(yscale="log")
    ax.set(xlabel="wavenumber")
    ax.legend()
    plt.tight_layout()
    return fig



# flake8: noqa
# Copied from https://github.com/NVIDIA/modulus/commit/89a6091bd21edce7be4e0539cbd91507004faf08
# Copyright (c) 2023 - 2024 NVIDIA CORPORATION & AFFILIATES.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dataclasses
from typing import Literal

import torch as th
import torch.nn as nn

from .healpix_layers import HEALPixLayer

# DOWNSAMPLING BLOCKS


class MaxPool(nn.Module):
    """Wrapper for applying Max Pooling with HEALPix or other tensor data.

    This class wraps the `nn.MaxPool2d` class to handle tensor data with
    HEALPix or other geometry layers.
    """

    def __init__(
        self,
        pooling: int = 2,
        enable_nhwc: bool = False,
        enable_healpixpad: bool = False,
    ):
        """
        Args:
            pooling (int, optional): Pooling kernel size passed to geometry layer.
            enable_nhwc (bool, optional): Enable nhwc format, passed to wrapper.
            enable_healpixpad (bool, optional): If HEALPixPadding should be enabled, passed to wrapper.
        """
        super().__init__()
        self.maxpool = HEALPixLayer(
            layer=nn.MaxPool2d,
            kernel_size=pooling,
            enable_nhwc=enable_nhwc,
            enable_healpixpad=enable_healpixpad,
        )

    def forward(self, x: th.Tensor) -> th.Tensor:
        """Forward pass of the MaxPool.

        Args:
            x: The values to MaxPool.

        Returns:
            The MaxPooled values.
        """
        return self.maxpool(x)


class AvgPool(nn.Module):
    """Wrapper for applying Average Pooling with HEALPix or other tensor data.

    This class wraps the `nn.AvgPool2d` class to handle tensor data with
    HEALPix or other geometry layers.
    """

    def __init__(
        self,
        pooling: int = 2,
        enable_nhwc: bool = False,
        enable_healpixpad: bool = False,
    ):
        """
        Args:
            pooling (int, optional): Pooling kernel size passed to geometry layer.
            enable_nhwc (bool, optional): Enable nhwc format, passed to wrapper.
            enable_healpixpad (bool, optional): If HEALPixPadding should be enabled, passed to wrapper.
        """
        super().__init__()
        self.avgpool = HEALPixLayer(
            layer=nn.AvgPool2d,
            kernel_size=pooling,
            enable_nhwc=enable_nhwc,
            enable_healpixpad=enable_healpixpad,
        )

    def forward(self, x: th.Tensor) -> th.Tensor:
        """Forward pass of the AvgPool layer.

        Args:
            x: The values to average.

        Returns:
            The averaged values.
        """
        return self.avgpool(x)


@dataclasses.dataclass
class DownsamplingBlockConfig:
    """
    Configuration for the downsampling block.
    Generally, either a pooling block or a striding conv block.

    Parameters:
        block_type: Type of recurrent block, either "MaxPool" or "AvgPool"
        pooling: Pooling size
        enable_nhwc: Flag to enable NHWC data format, default is False.
        enable_healpixpad: Flag to enable HEALPix padding, default is False.

    """

    block_type: Literal["MaxPool", "AvgPool"]
    pooling: int = 2
    enable_nhwc: bool = False
    enable_healpixpad: bool = False

    def build(self) -> nn.Module:
        """
        Builds the recurrent block model.

        Returns:
            Recurrent block.
        """
        if self.block_type == "MaxPool":
            return MaxPool(
                pooling=self.pooling,
                enable_nhwc=self.enable_nhwc,
                enable_healpixpad=self.enable_healpixpad,
            )

        elif self.block_type == "AvgPool":
            return AvgPool(
                pooling=self.pooling,
                enable_nhwc=self.enable_nhwc,
                enable_healpixpad=self.enable_healpixpad,
            )
        else:
            raise ValueError(f"Unsupported block type: {self.block_type}")


@dataclasses.dataclass
class UpsamplingBlockConfig:
    """
    Configuration for the upsampling block.
    Should be compatible with torch.nn.Upsample block configs.
    Attributes:
        block_type (str) – the upsampling algorithm: one of 'nearest', 'linear', 'bilinear',
            'bicubic' and 'trilinear'. Default: 'nearest'.
        scale_factor (float or Tuple[float]) - multiplier for spatial size. Has to match input size
            if it is a tuple.
        align_corners (bool, optional) – if True, the corner pixels of the input and output tensors are aligned,
            and thus preserving the values at those pixels. Only has effect when mode is 'linear', 'bilinear',
            'bicubic', or 'trilinear'. Default: False
    """

    scale_factor: int
    block_type: Literal["nearest", "linear", "bilinear", "bicubic", "trilinear"] = (
        "bilinear"
    )
    align_corners: bool = False

    def build(self) -> nn.Module:
        """
        Builds the upsampling block model.
        Returns:
            Upsampling block.
        """
        if self.align_corners == False:
            return nn.Upsample(
                scale_factor=self.scale_factor,
                mode=self.block_type,
            )  # Interpolate will complain about the align_corners option being passed at all, bad API design IMO
        else:
            return nn.Upsample(
                scale_factor=self.scale_factor,
                mode=self.block_type,
                align_corners=self.align_corners,
            )


@dataclasses.dataclass
class CappedGELUConfig:
    """
    Configuration for the CappedGELU activation function.

    Parameters:
        cap_value: Cap value for the GELU function, default is 10.
        enable_nhwc: Flag to enable NHWC data format, default is False.
        enable_healpixpad: Flag to enable HEALPix padding, default is False.
    """

    cap_value: int = 10
    enable_nhwc: bool = False
    enable_healpixpad: bool = False

    def build(self) -> nn.Module:
        """
        Builds the CappedGELU activation function.

        Returns:
            CappedGELU activation function.
        """
        return CappedGELU(cap_value=self.cap_value)


class CappedGELU(nn.Module):
    """
    Implements a GELU with capped maximum value.

    Example
    -------
    >>> capped_gelu_func = modulus.models.layers.CappedGELU()
    >>> input = th.Tensor([[-2,-1],[0,1],[2,3]])
    >>> capped_gelu_func(input)
    tensor([[-0.0455, -0.1587],
            [ 0.0000,  0.8413],
            [ 1.0000,  1.0000]])

    """

    def __init__(self, cap_value=1.0, **kwargs):
        """
        Args:
            cap_value: Maximum that values will be capped at
            **kwargs: Keyword arguments to be passed to the `th.nn.GELU` function
        """

        super().__init__()
        self.add_module("gelu", th.nn.GELU(**kwargs))
        self.register_buffer("cap", th.tensor(cap_value, dtype=th.float32))

    def forward(self, inputs):
        x = self.gelu(inputs)
        # Convert cap to a scalar value for clamping (ignores grad)
        cap_value = self.cap.item()
        x = th.clamp(x, max=cap_value)
        return x



import dataclasses
import datetime
import logging
from collections.abc import Callable
from typing import Any

import dacite
import torch
from torch import nn

from fme.core.corrector.atmosphere import AtmosphereCorrectorConfig
from fme.core.corrector.registry import CorrectorABC
from fme.core.dataset.utils import encode_timestep
from fme.core.dataset_info import DatasetInfo
from fme.core.device import get_device
from fme.core.dicts import add_names
from fme.core.distributed import Distributed
from fme.core.normalizer import NetworkAndLossNormalizationConfig, StandardNormalizer
from fme.core.ocean import Ocean, OceanConfig
from fme.core.optimization import NullOptimization
from fme.core.packer import Packer
from fme.core.registry import CorrectorSelector, ModuleSelector
from fme.core.step.step import StepABC, StepConfigABC, StepSelector
from fme.core.typing_ import TensorDict, TensorMapping

DEFAULT_TIMESTEP = datetime.timedelta(hours=6)
DEFAULT_ENCODED_TIMESTEP = encode_timestep(DEFAULT_TIMESTEP)


@StepSelector.register("single_module")
@dataclasses.dataclass
class SingleModuleStepConfig(StepConfigABC):
    """
    Configuration for a single module stepper.

    Parameters:
        builder: The module builder.
        in_names: Names of input variables.
        out_names: Names of output variables.
        normalization: The normalization configuration.
        ocean: The ocean configuration.
        corrector: The corrector configuration.
        next_step_forcing_names: Names of forcing variables for the next timestep.
        crps_training: Unused, kept for backwards compatibility.
        residual_prediction: Whether to use residual prediction.
    """

    builder: ModuleSelector
    in_names: list[str]
    out_names: list[str]
    normalization: NetworkAndLossNormalizationConfig
    ocean: OceanConfig | None = None
    corrector: AtmosphereCorrectorConfig | CorrectorSelector = dataclasses.field(
        default_factory=lambda: AtmosphereCorrectorConfig()
    )
    next_step_forcing_names: list[str] = dataclasses.field(default_factory=list)
    crps_training: bool | None = None
    residual_prediction: bool = False

    def __post_init__(self):
        self.crps_training = None  # unused, kept for backwards compatibility
        for name in self.next_step_forcing_names:
            if name not in self.in_names:
                raise ValueError(
                    f"next_step_forcing_name '{name}' not in in_names: {self.in_names}"
                )
            if name in self.out_names:
                raise ValueError(
                    f"next_step_forcing_name is an output variable: '{name}'"
                )

    @property
    def n_ic_timesteps(self) -> int:
        return 1

    def get_state(self):
        return dataclasses.asdict(self)

    def get_loss_normalizer(
        self,
        extra_names: list[str] | None = None,
        extra_residual_scaled_names: list[str] | None = None,
    ) -> StandardNormalizer:
        if extra_names is None:
            extra_names = []
        if extra_residual_scaled_names is None:
            extra_residual_scaled_names = []
        return self.normalization.get_loss_normalizer(
            names=self._normalize_names + extra_names,
            residual_scaled_names=self.prognostic_names + extra_residual_scaled_names,
        )

    @classmethod
    def from_state(cls, state) -> "SingleModuleStepConfig":
        state = cls._remove_deprecated_keys(state)
        return dacite.from_dict(
            data_class=cls, data=state, config=dacite.Config(strict=True)
        )

    @property
    def _normalize_names(self):
        """Names of variables which require normalization. I.e. inputs/outputs."""
        return list(set(self.in_names).union(self.out_names))

    @property
    def input_names(self) -> list[str]:
        """
        Names of variables required as inputs to `step`,
        either in `input` or `next_step_input_data`.
        """
        if self.ocean is None:
            return self.in_names
        else:
            return list(set(self.in_names).union(self.ocean.forcing_names))

    def get_next_step_forcing_names(self) -> list[str]:
        """Names of input-only variables which come from the output timestep."""
        return self.next_step_forcing_names

    @property
    def diagnostic_names(self) -> list[str]:
        """Names of variables which are outputs only."""
        return list(set(self.out_names).difference(self.in_names))

    @property
    def output_names(self) -> list[str]:
        return self.out_names

    @property
    def next_step_input_names(self) -> list[str]:
        """Names of variables provided in next_step_input_data."""
        input_only_names = set(self.input_names).difference(self.output_names)
        if self.ocean is None:
            return list(input_only_names)
        return list(input_only_names.union(self.ocean.forcing_names))

    @property
    def loss_names(self) -> list[str]:
        return self.output_names

    def replace_ocean(self, ocean: OceanConfig | None):
        """
        Replace the ocean model with a new one.

        Args:
            ocean: The new ocean model configuration or None.
        """
        self.ocean = ocean

    def get_ocean(self) -> OceanConfig | None:
        return self.ocean

    @classmethod
    def _remove_deprecated_keys(cls, state: dict[str, Any]) -> dict[str, Any]:
        state_copy = state.copy()
        return state_copy

    def get_step(
        self,
        dataset_info: DatasetInfo,
        init_weights: Callable[[list[nn.Module]], None],
    ) -> "SingleModuleStep":
        logging.info("Initializing stepper from provided config")
        corrector = dataset_info.vertical_coordinate.build_corrector(
            config=self.corrector,
            gridded_operations=dataset_info.gridded_operations,
            timestep=dataset_info.timestep,
        )
        normalizer = self.normalization.get_network_normalizer(self._normalize_names)
        return SingleModuleStep(
            config=self,
            img_shape=dataset_info.img_shape,
            corrector=corrector,
            normalizer=normalizer,
            timestep=dataset_info.timestep,
            init_weights=init_weights,
        )

    def load(self):
        self.normalization.load()


class SingleModuleStep(StepABC):
    """
    Step class for a single pytorch module.
    """

    TIME_DIM = 1
    CHANNEL_DIM = -3

    def __init__(
        self,
        config: SingleModuleStepConfig,
        img_shape: tuple[int, int],
        corrector: CorrectorABC,
        normalizer: StandardNormalizer,
        timestep: datetime.timedelta,
        init_weights: Callable[[list[nn.Module]], None],
    ):
        """
        Args:
            config: The configuration.
            img_shape: Shape of domain as (n_lat, n_lon).
            corrector: The corrector to use at the end of each step.
            normalizer: The normalizer to use.
            timestep: Timestep of the model.
            init_weights: Function to initialize the weights of the module.
        """
        super().__init__()
        n_in_channels = len(config.in_names)
        n_out_channels = len(config.out_names)
        self.in_packer = Packer(config.in_names)
        self.out_packer = Packer(config.out_names)
        self._normalizer = normalizer
        if config.ocean is not None:
            self.ocean: Ocean | None = config.ocean.build(
                config.in_names, config.out_names, timestep
            )
        else:
            self.ocean = None
        self.module = config.builder.build(
            n_in_channels=n_in_channels,
            n_out_channels=n_out_channels,
            img_shape=img_shape,
        ).to(get_device())
        init_weights([self.module])
        self._img_shape = img_shape
        self._config = config
        self._no_optimization = NullOptimization()

        dist = Distributed.get_instance()
        self.module = dist.wrap_module(self.module)

        self._timestep = timestep

        self._corrector = corrector
        self.in_names = config.in_names
        self.out_names = config.out_names

    @property
    def config(self) -> SingleModuleStepConfig:
        return self._config

    @property
    def normalizer(self) -> StandardNormalizer:
        return self._normalizer

    @property
    def surface_temperature_name(self) -> str | None:
        if self._config.ocean is not None:
            return self._config.ocean.surface_temperature_name
        return None

    @property
    def ocean_fraction_name(self) -> str | None:
        if self._config.ocean is not None:
            return self._config.ocean.ocean_fraction_name
        return None

    @property
    def modules(self) -> nn.ModuleList:
        """
        Returns:
            A list of modules being trained.
        """
        return nn.ModuleList([self.module])

    def step(
        self,
        input: TensorMapping,
        next_step_input_data: TensorMapping,
        wrapper: Callable[[nn.Module], nn.Module] = lambda x: x,
    ) -> TensorDict:
        """
        Step the model forward one timestep given input data.

        Args:
            input: Mapping from variable name to tensor of shape
                [n_batch, n_lat, n_lon] containing denormalized data from the
                initial timestep. In practice this contains the ML inputs.
            next_step_input_data: Mapping from variable name to tensor of shape
                [n_batch, n_lat, n_lon] containing denormalized data from
                the output timestep. In practice this contains the necessary data
                at the output timestep for the ocean model and corrector.
            wrapper: Wrapper to apply over each nn.Module before calling.

        Returns:
            The denormalized output data at the next time step.
        """

        def network_call(input_norm: TensorDict) -> TensorDict:
            input_tensor = self.in_packer.pack(input_norm, axis=self.CHANNEL_DIM)
            output_tensor = wrapper(self.module)(input_tensor)
            return self.out_packer.unpack(output_tensor, axis=self.CHANNEL_DIM)

        return step_with_adjustments(
            input=input,
            next_step_input_data=next_step_input_data,
            network_calls=network_call,
            normalizer=self.normalizer,
            corrector=self._corrector,
            ocean=self.ocean,
            residual_prediction=self._config.residual_prediction,
            prognostic_names=self.prognostic_names,
        )

    def get_regularizer_loss(self):
        return torch.tensor(0.0)

    def get_state(self):
        """
        Returns:
            The state of the stepper.
        """
        return {
            "module": self.module.state_dict(),
        }

    def load_state(self, state: dict[str, Any]) -> None:
        """
        Load the state of the stepper.

        Args:
            state: The state to load.
        """
        module = state["module"]
        if "module.device_buffer" in module:
            # for backwards compatibility with old checkpoints
            del module["module.device_buffer"]
        self.module.load_state_dict(module)


def step_with_adjustments(
    input: TensorMapping,
    next_step_input_data: TensorMapping,
    network_calls: Callable[[TensorDict], TensorDict],
    normalizer: StandardNormalizer,
    corrector: CorrectorABC,
    ocean: Ocean | None,
    residual_prediction: bool,
    prognostic_names: list[str],
) -> TensorDict:
    """
    Step the model forward one timestep given input data.

    Args:
        input: Mapping from variable name to tensor of shape
            [n_batch, n_lat, n_lon] containing denormalized data from the
            initial timestep. In practice this contains the ML inputs.
        next_step_input_data: Mapping from variable name to tensor of shape
            [n_batch, n_lat, n_lon] containing denormalized data from
            the output timestep. In practice this contains the necessary data
            at the output timestep for the ocean model and corrector.
        network_calls: Callable[[TensorMapping], TensorDict] that takes a
            normalized input and returns a normalized output.
        normalizer: The normalizer to use.
        corrector: The corrector to use at the end of each step.
        ocean: The ocean model to use.
        residual_prediction: Whether to use residual prediction.
        prognostic_names: Names of prognostic variables.

    Returns:
        The denormalized output data at the next time step.
    """
    input_norm = normalizer.normalize(input)
    output_norm = network_calls(input_norm)
    if residual_prediction:
        output_norm = add_names(input_norm, output_norm, prognostic_names)
    output = normalizer.denormalize(output_norm)
    if corrector is not None:
        output = corrector(input, output, next_step_input_data)
    if ocean is not None:
        output = ocean(input, output, next_step_input_data)
    return output



from collections.abc import Mapping, Sequence
from typing import Any

from omegaconf import OmegaConf


def update_dict_with_dotlist(
    base: Mapping[str, Any], dotlist: Sequence[str] | None = None
) -> Mapping[str, Any]:
    """Update a dictionary with a dotlist of key-value pairs.

    Args:
        base: The dictionary to update.
        dotlist: A list of key-value pairs with dots in the keys indicating nesting.

    Returns:
        The updated dictionary.

    Note:
        Uses omegaconf.from_dotlist to parse the dotlist.

    Example:
        >>> base = {"a": 1, "b": {"c": 2}}
        >>> dotlist = ["b.d=3", "e=4"]
        >>> update_dict_with_dotlist(base, dotlist)
        {"a": 1, "b": {"c": 2, "d": 3}, "e": 4}
    """
    if dotlist:
        override = OmegaConf.from_dotlist(dotlist)
        merged = OmegaConf.merge(base, override)
        return OmegaConf.to_container(merged, resolve=True)
    else:
        return base



import abc
import logging
from collections.abc import Callable

from fme.ace.data_loading.batch_data import BatchData, PairedData, PrognosticState
from fme.ace.inference.data_writer import PairedDataWriter
from fme.core.generics.aggregator import InferenceAggregatorABC, InferenceLogs
from fme.core.generics.data import InferenceDataABC
from fme.core.generics.inference import get_record_to_wandb
from fme.core.generics.writer import NullDataWriter
from fme.core.timing import GlobalTimer


class DeriverABC(abc.ABC):
    """
    Abstract base class for processing data during dataset comparison.
    """

    @abc.abstractmethod
    def get_forward_data(
        self, data: BatchData, compute_derived_variables: bool = False
    ) -> BatchData: ...

    @property
    @abc.abstractmethod
    def n_ic_timesteps(self) -> int: ...


def run_dataset_comparison(
    aggregator: InferenceAggregatorABC[PairedData, PairedData],
    prediction_data: InferenceDataABC[PrognosticState, BatchData],
    target_data: InferenceDataABC[PrognosticState, BatchData],
    deriver: DeriverABC,
    writer: PairedDataWriter | NullDataWriter | None = None,
    record_logs: Callable[[InferenceLogs], None] | None = None,
):
    if record_logs is None:
        record_logs = get_record_to_wandb(label="inference")
    if writer is None:
        writer = NullDataWriter()

    timer = GlobalTimer.get_instance()
    timer.start("data_loading")
    i_time = 0
    n_windows = min(len(prediction_data.loader), len(target_data.loader))
    for i, (pred, target) in enumerate(zip(prediction_data.loader, target_data.loader)):
        timer.stop()
        if i_time == 0:
            with timer.context("aggregator"):
                logs = aggregator.record_initial_condition(
                    initial_condition=PairedData.from_batch_data(
                        prediction=prediction_data.initial_condition.as_batch_data(),
                        reference=target_data.initial_condition.as_batch_data(),
                    ),
                )
            with timer.context("wandb_logging"):
                record_logs(logs)

        forward_steps_in_memory = list(pred.data.values())[0].size(1) - 1
        logging.info(
            f"Inference: Processing window {i + 1} of {n_windows}"
            f" spanning {i_time} to {i_time + forward_steps_in_memory} steps."
        )
        pred = deriver.get_forward_data(pred, compute_derived_variables=True)
        target = deriver.get_forward_data(target, compute_derived_variables=True)
        paired_data = PairedData.from_batch_data(prediction=pred, reference=target)

        with timer.context("data_writer"):
            writer.append_batch(
                batch=paired_data,
            )
        with timer.context("aggregator"):
            logs = aggregator.record_batch(
                data=paired_data,
            )

        with timer.context("wandb_logging"):
            record_logs(logs)

        timer.start("data_loading")
        i_time += forward_steps_in_memory

    timer.stop()



import dataclasses

import torch.nn as nn

from fme.ace.models.healpix.healpix_decoder import UNetDecoderConfig
from fme.ace.models.healpix.healpix_encoder import UNetEncoderConfig
from fme.ace.models.healpix.healpix_recunet import HEALPixRecUNet
from fme.ace.registry.registry import ModuleConfig, ModuleSelector


@ModuleSelector.register("HEALPixRecUNet")
@dataclasses.dataclass
class HEALPixRecUNetBuilder(ModuleConfig):
    """
    Configuration for the HEALPixRecUNet architecture used in DLWP.

    Parameters:
        presteps: Number of pre-steps, by default 1.
        input_time_size: Input time dimension, by default 0.
        output_time_size: Output time dimension, by default 0.
        delta_time: Delta time interval, by default "6h".
        reset_cycle: Reset cycle interval, by default "24h".
        input_channels: Number of input channels, by default 8.
        output_channels: Number of output channels, by default 8.
        n_constants: Number of constant input channels, by default 2.
        decoder_input_channels: Number of input channels for the decoder, by default 1.
        enable_nhwc: Flag to enable NHWC data format, by default False.
        enable_healpixpad: Flag to enable HEALPix padding, by default False.
    """

    encoder: UNetEncoderConfig
    decoder: UNetDecoderConfig
    presteps: int = 1
    input_time_size: int = 0
    output_time_size: int = 0
    delta_time: str = "6h"
    reset_cycle: str = "24h"
    n_constants: int = 2
    decoder_input_channels: int = 1
    prognostic_variables: int = 7
    enable_nhwc: bool = False
    enable_healpixpad: bool = False

    def build(
        self,
        n_in_channels: int,
        n_out_channels: int,
        img_shape: tuple[int, int],
    ) -> nn.Module:
        """
        Builds the HEALPixRecUNet model.

        Args:
            n_in_channels: Number of input channels.
            n_out_channels: Number of output channels.
            img_shape: Shape of the input image.

        Returns:
            HEALPixRecUNet model.
        """
        # Construct the HEALPixRecUNet module here using the parameters
        return HEALPixRecUNet(
            encoder=self.encoder,
            decoder=self.decoder,
            input_channels=n_in_channels,
            output_channels=n_out_channels,
            prognostic_variables=self.prognostic_variables,
            n_constants=self.n_constants,
            decoder_input_channels=self.decoder_input_channels,
            input_time_size=self.input_time_size,
            output_time_size=self.output_time_size,
            delta_time=self.delta_time,
            reset_cycle=self.reset_cycle,
            presteps=self.presteps,
            enable_nhwc=self.enable_nhwc,
            enable_healpixpad=self.enable_healpixpad,
        )



import numpy as np
import torch
import xarray as xr

from fme.core.device import get_device
from fme.core.distributed import Distributed
from fme.core.gridded_ops import GriddedOperations
from fme.core.typing_ import TensorMapping

from .reduced_metrics import AreaWeightedReducedMetric, ReducedMetric


def get_gen_shape(gen_data: TensorMapping):
    for name in gen_data:
        return gen_data[name].shape


class MeanAggregator:
    """
    Aggregator for mean-reduced metrics.

    These are metrics such as means which reduce to a single float for each batch,
    and then can be averaged across batches to get a single float for the
    entire dataset. This is important because the aggregator uses the mean to combine
    metrics across batches and processors.
    """

    def __init__(
        self,
        gridded_operations: GriddedOperations,
        target_time: int = 1,
    ):
        self._gridded_operations = gridded_operations
        self._n_batches = 0
        self._loss = torch.tensor(0.0, device=get_device())
        self._target_time = target_time
        self._dist = Distributed.get_instance()

        device = get_device()
        self._variable_metrics: dict[str, ReducedMetric] = {}
        self._variable_metrics["weighted_rmse"] = AreaWeightedReducedMetric(
            device=device,
            compute_metric=self._gridded_operations.area_weighted_rmse_dict,
        )
        self._variable_metrics["weighted_bias"] = AreaWeightedReducedMetric(
            device=device,
            compute_metric=self._gridded_operations.area_weighted_mean_bias_dict,
        )
        self._variable_metrics["weighted_grad_mag_percent_diff"] = (
            AreaWeightedReducedMetric(
                device=device,
                compute_metric=self._gridded_operations.area_weighted_gradient_magnitude_percent_diff_dict,  # noqa: E501
            )
        )

    @torch.no_grad()
    def record_batch(
        self,
        target_data: TensorMapping,
        gen_data: TensorMapping,
        target_data_norm: TensorMapping,
        gen_data_norm: TensorMapping,
        loss: torch.Tensor = torch.tensor(np.nan),
        i_time_start: int = 0,
    ):
        self._loss += loss
        time_dim = 1
        time_len = gen_data[list(gen_data.keys())[0]].shape[time_dim]
        target_time = self._target_time - i_time_start
        if target_time >= 0 and time_len > target_time:
            target_snapshot = {}
            gen_snapshot = {}
            for name in gen_data.keys():
                target_snapshot[name] = target_data[name].select(
                    dim=time_dim, index=target_time
                )
                gen_snapshot[name] = gen_data[name].select(
                    dim=time_dim, index=target_time
                )
            for metric in self._variable_metrics.values():
                metric.record(
                    target=target_snapshot,
                    gen=gen_snapshot,
                )
            # only increment n_batches if we actually recorded a batch
            self._n_batches += 1

    def _get_data(self):
        if self._n_batches == 0:
            raise ValueError("No batches have been recorded.")
        data: dict[str, torch.Tensor] = {"loss": self._loss / self._n_batches}
        for name, metric in self._variable_metrics.items():
            metric_results = metric.get()  # TensorDict: {var_name: metric_data}
            for key in metric_results:
                data[f"{name}/{key}"] = metric_results[key] / self._n_batches
        for key in sorted(data.keys()):
            data[key] = float(self._dist.reduce_mean(data[key].detach()).cpu().numpy())
        return data

    @torch.no_grad()
    def get_logs(self, label: str):
        """
        Returns logs as can be reported to WandB.

        Args:
            label: Label to prepend to all log keys.
        """
        return {
            f"{label}/{key}": data for key, data in sorted(self._get_data().items())
        }

    @torch.no_grad()
    def get_dataset(self) -> xr.Dataset:
        data = self._get_data()
        data = {key.replace("/", "-"): data[key] for key in data}
        data_vars = {}
        for key, value in data.items():
            data_vars[key] = xr.DataArray(value)
        return xr.Dataset(data_vars=data_vars)



import dataclasses
import datetime
import logging
from collections.abc import Callable
from typing import Any

import dacite
import torch
from torch import nn

from fme.core.corrector.atmosphere import AtmosphereCorrectorConfig
from fme.core.corrector.registry import CorrectorABC
from fme.core.dataset_info import DatasetInfo
from fme.core.device import get_device
from fme.core.distributed import Distributed
from fme.core.normalizer import NetworkAndLossNormalizationConfig, StandardNormalizer
from fme.core.ocean import Ocean, OceanConfig
from fme.core.optimization import NullOptimization
from fme.core.packer import Packer
from fme.core.registry import CorrectorSelector, ModuleSelector
from fme.core.step.single_module import step_with_adjustments
from fme.core.step.step import StepABC, StepConfigABC, StepSelector
from fme.core.typing_ import TensorDict, TensorMapping


@StepSelector.register("separate_radiation")
@dataclasses.dataclass
class SeparateRadiationStepConfig(StepConfigABC):
    """
    Configuration for a separate radiation stepper.

    Parameters:
        builder: The module builder.
        radiation_builder: The radiation module builder.
        main_prognostic_names: Names of prognostic variables. These are provided
            as input to both the main and radiation models, and output by
            the main model.
        shared_forcing_names: Names of forcing variables.
        radiation_only_forcing_names: Names of forcing variables for the radiation
            model, in addition to the ones specified in `shared_forcing_names`.
        radiation_diagnostic_names: Names of diagnostic variables for the radiation
            model.
        main_diagnostic_names: Names of diagnostic variables for the main model.
        normalization: The normalization configuration.
        next_step_forcing_names: Names of forcing variables which come from
            the output timestep.
        ocean: The ocean configuration.
        corrector: The corrector configuration.
        detach_radiation: Whether to detach the output of the radiation model before
            passing it to the main model. The radiation outputs returned by
            .step() will not be detached.
        residual_prediction: Whether to use residual prediction.
    """

    builder: ModuleSelector
    radiation_builder: ModuleSelector
    main_prognostic_names: list[str]
    shared_forcing_names: list[str]
    radiation_only_forcing_names: list[str]
    radiation_diagnostic_names: list[str]
    main_diagnostic_names: list[str]
    normalization: NetworkAndLossNormalizationConfig
    next_step_forcing_names: list[str] = dataclasses.field(default_factory=list)
    ocean: OceanConfig | None = None
    corrector: AtmosphereCorrectorConfig | CorrectorSelector = dataclasses.field(
        default_factory=lambda: AtmosphereCorrectorConfig()
    )
    detach_radiation: bool = False
    residual_prediction: bool = False

    def __post_init__(self):
        seen_names: dict[str, str] = {}
        for name_list, label in (
            (self.main_prognostic_names, "main_prognostic_names"),
            (self.shared_forcing_names, "shared_forcing_names"),
            (self.radiation_only_forcing_names, "radiation_only_forcing_names"),
            (self.main_diagnostic_names, "main_diagnostic_names"),
            (self.radiation_diagnostic_names, "radiation_diagnostic_names"),
        ):
            for name in name_list:
                if name in seen_names:
                    raise ValueError(
                        f"Name '{name}' appears in multiple name lists: "
                        f"{seen_names[name]} and {label}."
                    )
            seen_names[name] = label
        for name in self.next_step_forcing_names:
            if name not in self._forcing_names:
                raise ValueError(
                    "next_step_forcing_name not in forcing_names: "
                    f"'{name}' not in {self._forcing_names}"
                )

    @property
    def n_ic_timesteps(self) -> int:
        return 1

    def get_state(self):
        return dataclasses.asdict(self)

    def get_step(
        self,
        dataset_info: DatasetInfo,
        init_weights: Callable[[list[nn.Module]], None],
    ) -> "SeparateRadiationStep":
        logging.info("Initializing stepper from provided config")
        corrector = dataset_info.vertical_coordinate.build_corrector(
            config=self.corrector,
            gridded_operations=dataset_info.gridded_operations,
            timestep=dataset_info.timestep,
        )
        normalizer = self.normalization.get_network_normalizer(self._normalize_names)
        return SeparateRadiationStep(
            config=self,
            img_shape=dataset_info.img_shape,
            corrector=corrector,
            normalizer=normalizer,
            timestep=dataset_info.timestep,
            init_weights=init_weights,
        )

    def get_loss_normalizer(
        self,
        extra_names: list[str] | None = None,
        extra_residual_scaled_names: list[str] | None = None,
    ) -> StandardNormalizer:
        if extra_names is None:
            extra_names = []
        if extra_residual_scaled_names is None:
            extra_residual_scaled_names = []
        return self.normalization.get_loss_normalizer(
            names=self._normalize_names + extra_names,
            residual_scaled_names=self.prognostic_names + extra_residual_scaled_names,
        )

    @classmethod
    def from_state(cls, state) -> "SeparateRadiationStepConfig":
        return dacite.from_dict(
            data_class=cls, data=state, config=dacite.Config(strict=True)
        )

    @property
    def _normalize_names(self) -> list[str]:
        """Names of variables which require normalization. I.e. inputs/outputs."""
        all_names = set()
        for names in (
            self.main_prognostic_names,
            self.shared_forcing_names,
            self.radiation_only_forcing_names,
            self.main_diagnostic_names,
            self.radiation_diagnostic_names,
        ):
            all_names.update(names)
        return list(all_names)

    @property
    def _forcing_names(self) -> list[str]:
        return list(
            set(self.shared_forcing_names).union(self.radiation_only_forcing_names)
        )

    def get_next_step_forcing_names(self) -> list[str]:
        return self.next_step_forcing_names

    @property
    def diagnostic_names(self) -> list[str]:
        return list(
            set(self.main_diagnostic_names).union(self.radiation_diagnostic_names)
        )

    @property
    def radiation_in_names(self) -> list[str]:
        return (
            self.main_prognostic_names
            + self.shared_forcing_names
            + self.radiation_only_forcing_names
        )

    @property
    def radiation_out_names(self) -> list[str]:
        return self.radiation_diagnostic_names

    @property
    def main_in_names(self) -> list[str]:
        return (
            self.main_prognostic_names
            + self.shared_forcing_names
            + self.radiation_out_names
        )

    @property
    def main_out_names(self) -> list[str]:
        return self.main_prognostic_names + self.main_diagnostic_names

    @property
    def input_names(self) -> list[str]:
        ml_in_names = (
            self.main_prognostic_names
            + self.shared_forcing_names
            + self.radiation_only_forcing_names
        )
        if self.ocean is None:
            return ml_in_names
        else:
            return list(set(ml_in_names).union(self.ocean.forcing_names))

    @property
    def output_names(self) -> list[str]:
        return (
            self.main_prognostic_names
            + self.main_diagnostic_names
            + self.radiation_diagnostic_names
        )

    @property
    def next_step_input_names(self) -> list[str]:
        """Names of variables provided in next_step_input_data."""
        input_only_names = set(self.input_names).difference(self.output_names)
        if self.ocean is None:
            return list(input_only_names)
        return list(input_only_names.union(self.ocean.forcing_names))

    @property
    def loss_names(self) -> list[str]:
        return self.output_names

    def replace_ocean(self, ocean: OceanConfig | None):
        self.ocean = ocean

    def get_ocean(self) -> OceanConfig | None:
        return self.ocean

    def load(self):
        self.normalization.load()


class SeparateRadiationStep(StepABC):
    """
    Step class for a single pytorch module.
    """

    TIME_DIM = 1
    CHANNEL_DIM = -3

    def __init__(
        self,
        config: SeparateRadiationStepConfig,
        img_shape: tuple[int, int],
        corrector: CorrectorABC,
        normalizer: StandardNormalizer,
        timestep: datetime.timedelta,
        init_weights: Callable[[list[nn.Module]], None],
    ):
        """
        Args:
            config: The configuration.
            img_shape: Shape of domain as (n_lat, n_lon).
            corrector: The corrector to use at the end of each step.
            normalizer: The normalizer to use.
            timestep: Timestep of the model.
            init_weights: Function to initialize the weights of the step.
        """
        super().__init__()
        self.in_packer = Packer(config.main_in_names)
        self.out_packer = Packer(config.main_out_names)
        self.radiation_in_packer = Packer(config.radiation_in_names)
        self.radiation_out_packer = Packer(config.radiation_out_names)
        self._normalizer = normalizer
        if config.ocean is not None:
            self.ocean: Ocean | None = config.ocean.build(
                config.input_names,
                config.output_names,
                timestep,
            )
        else:
            self.ocean = None
        self.module: nn.Module = config.builder.build(
            n_in_channels=len(config.main_in_names),
            n_out_channels=len(config.main_out_names),
            img_shape=img_shape,
        ).to(get_device())
        self.radiation_module: nn.Module = config.radiation_builder.build(
            n_in_channels=len(config.radiation_in_names),
            n_out_channels=len(config.radiation_out_names),
            img_shape=img_shape,
        ).to(get_device())
        self._img_shape = img_shape
        self._config = config
        self._no_optimization = NullOptimization()

        init_weights(self.modules)
        dist = Distributed.get_instance()
        self.module = dist.wrap_module(self.module)
        self.radiation_module = dist.wrap_module(self.radiation_module)
        self._timestep = timestep
        self._corrector = corrector

    @property
    def config(self) -> SeparateRadiationStepConfig:
        return self._config

    @property
    def surface_temperature_name(self) -> str | None:
        if self._config.ocean is not None:
            return self._config.ocean.surface_temperature_name
        return None

    @property
    def ocean_fraction_name(self) -> str | None:
        if self._config.ocean is not None:
            return self._config.ocean.ocean_fraction_name
        return None

    @property
    def normalizer(self) -> StandardNormalizer:
        return self._normalizer

    @property
    def modules(self) -> nn.ModuleList:
        """
        Returns:
            A list of modules being trained.
        """
        return nn.ModuleList([self.module, self.radiation_module])

    def step(
        self,
        input: TensorMapping,
        next_step_forcing_data: TensorMapping,
        wrapper: Callable[[nn.Module], nn.Module] = lambda x: x,
    ) -> TensorDict:
        """
        Step the model forward one timestep given input data.

        Args:
            input: Mapping from variable name to tensor of shape
                [n_batch, n_lat, n_lon]. This data is used as input for `self.module`
                and is assumed to contain all input variables and be denormalized.
            next_step_forcing_data: Mapping from variable name to tensor of shape
                [n_batch, n_lat, n_lon]. This must contain the necessary forcing
                data at the output timestep for the ocean model and corrector.
            wrapper: Wrapper to apply over each nn.Module before calling.

        Returns:
            The denormalized output data at the next time step.
        """

        def network_calls(input_norm: TensorDict) -> TensorDict:
            radiation_input_tensor = self.radiation_in_packer.pack(
                input_norm, axis=self.CHANNEL_DIM
            )
            radiation_output_tensor = wrapper(self.radiation_module)(
                radiation_input_tensor
            )
            radiation_output_norm = self.radiation_out_packer.unpack(
                radiation_output_tensor, axis=self.CHANNEL_DIM
            )
            main_input_data = input_norm.copy()
            if self._config.detach_radiation:
                main_input_data = {
                    **input_norm,
                    **{k: v.detach() for k, v in radiation_output_norm.items()},
                }
            else:
                main_input_data = {**input_norm, **radiation_output_norm}
            input_tensor = self.in_packer.pack(main_input_data, axis=self.CHANNEL_DIM)
            output_tensor = wrapper(self.module)(input_tensor)
            main_output_norm = self.out_packer.unpack(
                output_tensor, axis=self.CHANNEL_DIM
            )
            return {
                **radiation_output_norm,
                **main_output_norm,
            }

        return step_with_adjustments(
            input=input,
            next_step_input_data=next_step_forcing_data,
            network_calls=network_calls,
            normalizer=self.normalizer,
            corrector=self._corrector,
            ocean=self.ocean,
            residual_prediction=self._config.residual_prediction,
            prognostic_names=self.prognostic_names,
        )

    def get_regularizer_loss(self) -> torch.Tensor:
        return torch.tensor(0.0)

    def get_state(self):
        """
        Returns:
            The state of the ML modules.
        """
        return {
            "module": self.module.state_dict(),
            "radiation_module": self.radiation_module.state_dict(),
        }

    def load_state(self, state: dict[str, Any]) -> None:
        """
        Load the state of the ML modules.

        Args:
            state: The state to load.
        """
        self.module.load_state_dict(state["module"])
        self.radiation_module.load_state_dict(state["radiation_module"])



from .inference import InferenceEvaluatorAggregator, InferenceEvaluatorAggregatorConfig
from .null import NullAggregator
from .one_step import OneStepAggregator
from .one_step.main import OneStepAggregatorConfig
from .train import TrainAggregator



import dataclasses
from collections.abc import Mapping
from typing import NewType

import torch

TensorMapping = Mapping[str, torch.Tensor]
TensorDict = dict[str, torch.Tensor]
EnsembleTensorDict = NewType("EnsembleTensorDict", TensorDict)
EnsembleTensorDict.__doc__ = """
A dictionary of tensors with an explicit ensemble (sample) dimension, where
ensemble members represent multiple predictions for the same initial condition.

The ensemble dimension is the second dimension of the tensors,
while the batch dimension is the first.
"""


@dataclasses.dataclass
class Slice:
    """
    Configuration of a python `slice` built-in.

    Required because `slice` cannot be initialized directly by dacite.

    Parameters:
        start: Start index of the slice.
        stop: Stop index of the slice.
        step: Step of the slice.
    """

    start: int | None = None
    stop: int | None = None
    step: int | None = None

    @property
    def slice(self) -> slice:
        return slice(self.start, self.stop, self.step)

    def contains(self, value: int) -> bool:
        start = self.start if self.start is not None else 0
        stop = self.stop if self.stop is not None else float("inf")
        step = self.step if self.step is not None else 1
        return start <= value < stop and (value - start) % step == 0



import abc
import dataclasses
from collections.abc import Callable, Mapping

# we use Type to distinguish from type attr of ModuleSelector
from typing import Any, ClassVar, Type  # noqa: UP035

import dacite
from torch import nn

from .registry import Registry


@dataclasses.dataclass
class ModuleConfig(abc.ABC):
    """
    Builds a nn.Module given information about the input
    and output channels and the image shape.

    This is a "Config" as in practice it is a dataclass loaded directly from yaml,
    allowing us to specify details of the network architecture in a config file.
    """

    @abc.abstractmethod
    def build(
        self,
        n_in_channels: int,
        n_out_channels: int,
        img_shape: tuple[int, int],
    ) -> nn.Module:
        """
        Build a nn.Module given information about the input and output channels
        and the image shape.

        Args:
            n_in_channels: number of input channels
            n_out_channels: number of output channels
            img_shape: shape of last two dimensions of data, e.g. latitude and
                longitude.

        Returns:
            a nn.Module
        """
        ...

    @classmethod
    def from_state(cls, state: Mapping[str, Any]) -> "ModuleConfig":
        """
        Create a ModuleSelector from a dictionary containing all the information
        needed to build a ModuleConfig.
        """
        return dacite.from_dict(
            data_class=cls, data=state, config=dacite.Config(strict=True)
        )


@dataclasses.dataclass
class ModuleSelector:
    """
    A dataclass containing all the information needed to build a ModuleConfig,
    including the type of the ModuleConfig and the data needed to build it.

    This is helpful as ModuleSelector can be serialized and deserialized
    without any additional information, whereas to load a ModuleConfig you
    would need to know the type of the ModuleConfig being loaded.

    It is also convenient because ModuleSelector is a single class that can be
    used to represent any ModuleConfig, whereas ModuleConfig is a protocol
    that can be implemented by many different classes.

    Parameters:
        type: the type of the ModuleConfig
        config: data for a ModuleConfig instance of the indicated type
    """

    type: str
    config: Mapping[str, Any]
    registry: ClassVar[Registry[ModuleConfig]] = Registry[ModuleConfig]()

    def __post_init__(self):
        if not isinstance(self.registry, Registry):
            raise ValueError("ModuleSelector.registry should not be set manually")
        self._instance = self.registry.get(self.type, self.config)

    @classmethod
    def register(
        cls, type_name: str
    ) -> Callable[[Type[ModuleConfig]], Type[ModuleConfig]]:  # noqa: UP006
        return cls.registry.register(type_name)

    def build(
        self,
        n_in_channels: int,
        n_out_channels: int,
        img_shape: tuple[int, int],
    ) -> nn.Module:
        """
        Build a nn.Module given information about the input and output channels
        and the image shape.

        Args:
            n_in_channels: number of input channels
            n_out_channels: number of output channels
            img_shape: shape of last two dimensions of data, e.g. latitude and
                longitude.

        Returns:
            a nn.Module
        """
        return self._instance.build(
            n_in_channels=n_in_channels,
            n_out_channels=n_out_channels,
            img_shape=img_shape,
        )

    @classmethod
    def get_available_types(cls):
        """This class method is used to expose all available types of Modules."""
        module = nn.Identity()
        return cls(type="prebuilt", config={"module": module}).registry._types.keys()



import dataclasses
from collections.abc import Callable, Collection, Sequence
from typing import Any, TypeVar

import numpy as np
import torch
import xarray as xr
from torch.utils.data import default_collate

from fme.core.device import get_device
from fme.core.typing_ import TensorDict, TensorMapping

SelfType = TypeVar("SelfType", bound="BatchData")


def _check_device(data: TensorMapping, device: torch.device):
    for v in data.values():
        if v.device != device:
            raise ValueError(f"data must be on {device}")


class PrognosticState:
    """
    Thin typing wrapper around BatchData to indicate that the data is a prognostic
    state, such as an initial condition or final state when evolving forward in time.
    """

    def __init__(self, data: "BatchData"):
        """
        Initialize the state.

        Args:
            data: The data to initialize the state with.
        """
        self._data = data

    def to_device(self) -> "PrognosticState":
        return PrognosticState(self._data.to_device())

    def as_batch_data(self) -> "BatchData":
        return self._data


@dataclasses.dataclass
class BatchData:
    """A container for the data and time coordinates of a batch.

    Parameters:
        data: Data for each variable in each sample of shape (sample, time, ...),
            concatenated along samples to make a batch. To be used directly in training,
            validation, and inference.
        time: An array representing time coordinates for each sample in the batch,
            concatenated along samples to make a batch. To be used in writing out
            inference predictions with time coordinates, not directly in ML.
        horizontal_dims: Horizontal dimensions of the data. Used for writing to
            netCDF files.
    """

    data: TensorMapping
    time: xr.DataArray
    horizontal_dims: list[str] = dataclasses.field(
        default_factory=lambda: ["lat", "lon"]
    )

    @property
    def dims(self) -> list[str]:
        return ["sample", "time"] + self.horizontal_dims

    @property
    def n_timesteps(self) -> int:
        return self.time["time"].values.size

    def to_device(self) -> "BatchData":
        return self.__class__(
            data={k: v.to(get_device()) for k, v in self.data.items()},
            time=self.time,
            horizontal_dims=self.horizontal_dims,
        )

    @classmethod
    def _get_kwargs(cls, horizontal_dims: list[str] | None) -> dict[str, Any]:
        if horizontal_dims is None:
            kwargs = {}
        else:
            kwargs = {"horizontal_dims": horizontal_dims}
        return kwargs

    @classmethod
    def new_on_cpu(
        cls,
        data: TensorMapping,
        time: xr.DataArray,
        horizontal_dims: list[str] | None = None,
    ) -> "BatchData":
        _check_device(data, torch.device("cpu"))
        kwargs = cls._get_kwargs(horizontal_dims)
        return BatchData(
            data=data,
            time=time,
            **kwargs,
        )

    @classmethod
    def new_on_device(
        cls,
        data: TensorMapping,
        time: xr.DataArray,
        horizontal_dims: list[str] | None = None,
    ) -> "BatchData":
        """
        Move the data to the current global device specified by get_device().
        """
        _check_device(data, get_device())
        kwargs = cls._get_kwargs(horizontal_dims)
        return BatchData(
            data=data,
            time=time,
            **kwargs,
        )

    def __post_init__(self):
        if len(self.time.shape) != 2:
            raise ValueError(
                "Expected time to have shape (n_samples, n_times), got shape "
                f"{self.time.shape}."
            )
        for k, v in self.data.items():
            if v.shape[:2] != self.time.shape[:2]:
                raise ValueError(
                    f"Data for variable {k} has shape {v.shape}, expected shape "
                    f"(n_samples, n_times) for time but got shape "
                    f"{self.time.shape}."
                )

    @classmethod
    def from_sample_tuples(
        cls,
        samples: Sequence[tuple[TensorMapping, xr.DataArray]],
        sample_dim_name: str = "sample",
        horizontal_dims: list[str] | None = None,
    ) -> "BatchData":
        sample_data, sample_times = zip(*samples)
        batch_data = default_collate(sample_data)
        batch_time = xr.concat(sample_times, dim=sample_dim_name)
        return BatchData.new_on_cpu(
            data=batch_data,
            time=batch_time,
            horizontal_dims=horizontal_dims,
        )

    def compute_derived_variables(
        self: SelfType,
        derive_func: Callable[[TensorMapping, TensorMapping], TensorDict],
        forcing_data: SelfType,
    ) -> SelfType:
        """
        Compute derived variables from the data and forcing data.

        The forcing data must have the same time coordinate as the batch data.

        Args:
            derive_func: A function that takes the data and forcing data and returns a
                dictionary of derived variables.
            forcing_data: The forcing data to compute derived variables from.
        """
        if not np.all(forcing_data.time.values == self.time.values):
            raise ValueError(
                "Forcing data must have the same time coordinate as the batch data."
            )
        derived_data = derive_func(self.data, forcing_data.data)
        return self.__class__(
            data={**self.data, **derived_data},
            time=self.time,
            horizontal_dims=self.horizontal_dims,
        )

    def remove_initial_condition(self: SelfType, n_ic_timesteps: int) -> SelfType:
        """
        Remove the initial condition timesteps from the data.
        """
        if n_ic_timesteps == 0:
            raise RuntimeError("No initial condition timesteps to remove.")
        return self.__class__(
            {k: v[:, n_ic_timesteps:] for k, v in self.data.items()},
            time=self.time.isel(time=slice(n_ic_timesteps, None)),
            horizontal_dims=self.horizontal_dims,
        )

    def subset_names(self: SelfType, names: Collection[str]) -> SelfType:
        """
        Subset the data to only include the given names.
        """
        return self.__class__(
            {k: v for k, v in self.data.items() if k in names},
            time=self.time,
            horizontal_dims=self.horizontal_dims,
        )

    def get_start(
        self: SelfType, prognostic_names: Collection[str], n_ic_timesteps: int
    ) -> PrognosticState:
        """
        Get the initial condition state.
        """
        return PrognosticState(
            self.subset_names(prognostic_names).select_time_slice(
                slice(0, n_ic_timesteps)
            )
        )

    def get_end(
        self: SelfType, prognostic_names: Collection[str], n_ic_timesteps: int
    ) -> PrognosticState:
        """
        Get the final state which can be used as a new initial condition.
        """
        return PrognosticState(
            self.subset_names(prognostic_names).select_time_slice(
                slice(-n_ic_timesteps, None)
            )
        )

    def select_time_slice(self: SelfType, time_slice: slice) -> SelfType:
        """
        Select a window of data from the batch.
        """
        return self.__class__(
            {k: v[:, time_slice] for k, v in self.data.items()},
            time=self.time[:, time_slice],
            horizontal_dims=self.horizontal_dims,
        )

    def prepend(self: SelfType, initial_condition: PrognosticState) -> SelfType:
        """
        Prepend the initial condition to the data.
        """
        initial_batch_data = initial_condition.as_batch_data()
        filled_data = {**initial_batch_data.data}
        example_tensor = list(initial_batch_data.data.values())[0]
        state_data_device = list(self.data.values())[0].device
        for k in self.data:
            if k not in filled_data:
                filled_data[k] = torch.full_like(example_tensor, fill_value=np.nan)
        return self.__class__(
            data={
                k: torch.cat([filled_data[k].to(state_data_device), v], dim=1)
                for k, v in self.data.items()
            },
            time=xr.concat([initial_batch_data.time, self.time], dim="time"),
            horizontal_dims=self.horizontal_dims,
        )


@dataclasses.dataclass
class PairedData:
    """A container for the data and time coordinate of a batch, with paired
    prediction and target data.
    """

    prediction: TensorMapping
    reference: TensorMapping
    time: xr.DataArray

    @property
    def forcing(self) -> TensorMapping:
        return {k: v for k, v in self.reference.items() if k not in self.prediction}

    @property
    def target(self) -> TensorMapping:
        return {k: v for k, v in self.reference.items() if k in self.prediction}

    @classmethod
    def from_batch_data(
        cls,
        prediction: BatchData,
        reference: BatchData,
    ) -> "PairedData":
        if not np.all(prediction.time.values == reference.time.values):
            raise ValueError("Prediction and target time coordinate must be the same.")
        return PairedData(
            prediction=prediction.data, reference=reference.data, time=prediction.time
        )

    @classmethod
    def new_on_device(
        cls,
        prediction: TensorMapping,
        reference: TensorMapping,
        time: xr.DataArray,
    ) -> "PairedData":
        device = get_device()
        _check_device(prediction, device)
        _check_device(reference, device)
        return PairedData(prediction, reference, time)

    @classmethod
    def new_on_cpu(
        cls,
        prediction: TensorMapping,
        reference: TensorMapping,
        time: xr.DataArray,
    ) -> "PairedData":
        _check_device(prediction, torch.device("cpu"))
        _check_device(reference, torch.device("cpu"))
        return PairedData(prediction, reference, time)



import dataclasses

from .registry import Registry


def test_registry_without_base_type():
    """Test that Registry without a base type works."""

    # Create registry that accepts any class
    model_registry: Registry = Registry()

    # Register a class
    @model_registry.register("model_a")
    @dataclasses.dataclass
    class ModelA:
        pass

    assert isinstance(model_registry.get("model_a", {}), ModelA)


def test_registry_type_checking():
    """Test that Registry enforces type checking on registered classes."""

    # Base class for type checking
    class BaseModel:
        pass

    # Create registry that only accepts BaseModel subclasses
    model_registry: Registry[BaseModel] = Registry[BaseModel]()

    # This should work - ModelA inherits from BaseModel
    @model_registry.register("model_a")
    @dataclasses.dataclass
    class ModelA(BaseModel):
        pass

    try:
        # This would fail in mypy as ModelB is not a subclass of BaseModel
        @model_registry.register("model_b")  # type: ignore[arg-type]
        @dataclasses.dataclass
        class ModelB:
            pass

        assert False, "Expected TypeError - ModelB is not a subclass of BaseModel"
    except TypeError:
        pass



import logging

import numpy as np
import torch
import xarray as xr

from fme.core.dataset.xarray import XarrayDataset
from fme.core.typing_ import TensorDict


class XarraySubset(torch.utils.data.Dataset):
    def __init__(self, dataset: XarrayDataset, subset: slice | np.ndarray):
        indices = np.arange(len(dataset))[subset]
        logging.info(f"Subsetting dataset samples according to {subset}.")
        self._dataset = torch.utils.data.Subset(dataset, indices)
        self._sample_start_times = dataset.sample_start_times[indices]
        self._sample_n_times = dataset.sample_n_times
        self.dims = dataset.dims

    def __len__(self):
        return len(self._dataset)

    def __getitem__(self, idx: int) -> tuple[TensorDict, xr.DataArray]:
        return self._dataset[idx]

    @property
    def sample_start_times(self):
        return self._sample_start_times

    @property
    def sample_n_times(self) -> int:
        """The length of the time dimension of each sample."""
        return self._sample_n_times
